import base64
import hashlib
import io
import tempfile
import time
from typing import Annotated, Optional

from fastapi import APIRouter, Header, Request, Response
from lonelypsp.auth.config import AuthResult
from lonelypsp.compat import assert_never
from lonelypsp.stateless.constants import SubscriberToBroadcasterStatelessMessageType
from lonelypsp.sync_io import PreallocatedBytesIO

from lonelypss.middleware.config import get_config_from_request
from lonelypss.middleware.ws_receiver import get_ws_receiver_from_request
from lonelypss.util.async_io import async_read_exact
from lonelypss.util.request_body_io import AsyncIterableAIO

router = APIRouter()


@router.post(
    "/v1/receive_for_websockets",
)
async def receive_for_websockets(
    request: Request,
    authorization: Annotated[Optional[str], Header()] = None,
    repr_digest: Annotated[Optional[str], Header()] = None,
) -> Response:
    """As a broadcaster, in order to handle websocket connections, we need to be notified
    about messages that were sent to other broadcasters. To facilitate this, the broadcaster
    acts as a subscriber for itself, using this endpoint to receive messages, then forwards
    these to an in-memory structure to fan it out to all the websocket connections.

    This is not the endpoint that subscribers use to notify broadcasters. Use `/v1/notify` for
    that.
    """
    config = get_config_from_request(request)
    receiver = get_ws_receiver_from_request(request)

    if repr_digest is None:
        return Response(
            status_code=400,
            headers={"Content-Type": "application/octet-stream"},
            content=int(
                SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
            ).to_bytes(2, "big"),
        )

    expected_digest_b64: Optional[str] = None
    for digest_pair in repr_digest.split(","):
        split_digest_pair = digest_pair.split("=", 1)
        if len(split_digest_pair) != 2:
            continue
        digest_type, digest_value = split_digest_pair
        if digest_type != "sha-512":
            continue

        expected_digest_b64 = digest_value

    if expected_digest_b64 is None:
        return Response(
            status_code=400,
            headers={"Content-Type": "application/octet-stream"},
            content=int(
                SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
            ).to_bytes(2, "big"),
        )

    try:
        expected_digest = base64.b64decode(expected_digest_b64 + "==")
    except BaseException:
        return Response(
            status_code=400,
            headers={"Content-Type": "application/octet-stream"},
            content=int(
                SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
            ).to_bytes(2, "big"),
        )

    request_url = str(request.url)
    if request_url != receiver.receiver_url:
        return Response(
            status_code=400,
            headers={"Content-Type": "application/octet-stream"},
            content=int(
                SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
            ).to_bytes(2, "big"),
        )

    stream = request.stream()
    try:
        body = AsyncIterableAIO(stream)

        try:
            tracing_length_bytes = await async_read_exact(body, 2)
            tracing_length = int.from_bytes(tracing_length_bytes, "big")
            tracing = await async_read_exact(body, tracing_length)

            topic_length_bytes = await async_read_exact(body, 2)
            topic_length = int.from_bytes(topic_length_bytes, "big")
            topic = await async_read_exact(body, topic_length)

            identifier_length_bytes = await async_read_exact(body, 1)
            identifier_length = int.from_bytes(identifier_length_bytes, "big")
            identifier = await async_read_exact(body, identifier_length)

            message_length_bytes = await async_read_exact(body, 8)
            message_length = int.from_bytes(message_length_bytes, "big")
        except ValueError:
            return Response(status_code=400)

        if not receiver.is_relevant(topic):
            return Response(
                status_code=400,
                headers={"Content-Type": "application/octet-stream"},
                content=int(
                    SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
                ).to_bytes(2, "big"),
            )

        auth_result = await config.is_receive_allowed(
            tracing=tracing,
            url=str(request.url),
            topic=topic,
            message_sha512=expected_digest,
            identifier=identifier,
            now=time.time(),
            authorization=authorization,
        )
        if auth_result == AuthResult.UNAVAILABLE:
            return Response(status_code=503)
        if auth_result == AuthResult.FORBIDDEN:
            return Response(
                status_code=403,
                headers={"Content-Type": "application/octet-stream"},
                content=int(
                    SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
                ).to_bytes(2, "big"),
            )
        if auth_result == AuthResult.UNAUTHORIZED:
            return Response(
                status_code=401,
                headers={"Content-Type": "application/octet-stream"},
                content=int(
                    SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
                ).to_bytes(2, "big"),
            )
        if auth_result != AuthResult.OK:
            assert_never(auth_result)

        with tempfile.SpooledTemporaryFile(
            max_size=config.message_body_spool_size, mode="w+b"
        ) as spooled_request_body:
            read_length = 0
            hasher = hashlib.sha512()
            while True:
                chunk = await body.read(io.DEFAULT_BUFFER_SIZE)
                if not chunk:
                    break
                hasher.update(chunk)
                read_length += len(chunk)
                spooled_request_body.write(chunk)

                if read_length > message_length:
                    return Response(status_code=400)

            await stream.aclose()
            if read_length != message_length:
                return Response(status_code=400)

            real_digest = hasher.digest()
            if real_digest != expected_digest:
                return Response(status_code=400)

            spooled_request_body.seek(0)
            if read_length < config.message_body_spool_size:
                small_body = spooled_request_body.read()
                count = await receiver.on_small_incoming(
                    small_body, topic=topic, sha512=real_digest
                )
            else:
                count = await receiver.on_large_exclusive_incoming(
                    spooled_request_body,
                    topic=topic,
                    sha512=real_digest,
                    length=read_length,
                )

        resp_tracing = b""  # TODO: tracing
        resp_authorization = await config.authorize_confirm_receive(
            tracing=resp_tracing,
            identifier=identifier,
            num_subscribers=count,
            url=str(request.url),
            now=time.time(),
        )
        resp_authorization_bytes = (
            b"" if resp_authorization is None else resp_authorization.encode("utf-8")
        )
        resp_body = PreallocatedBytesIO(
            2
            + 2
            + len(resp_authorization_bytes)
            + 2
            + len(resp_tracing)
            + 1
            + len(identifier)
            + 4
        )
        resp_body.write(
            int(
                SubscriberToBroadcasterStatelessMessageType.RESPONSE_CONFIRM_RECEIVE
            ).to_bytes(2, "big")
        )
        resp_body.write(len(resp_authorization_bytes).to_bytes(2, "big"))
        resp_body.write(resp_authorization_bytes)
        resp_body.write(len(resp_tracing).to_bytes(2, "big"))
        resp_body.write(resp_tracing)
        resp_body.write(len(identifier).to_bytes(1, "big"))
        resp_body.write(identifier)
        resp_body.write(count.to_bytes(4, "big"))
        return Response(
            status_code=200,
            headers={"Content-Type": "application/octet-stream"},
            content=memoryview(resp_body.buffer),
        )
    finally:
        await stream.aclose()



import asyncio
import io
import tempfile
from typing import Union

from lonelypsp.util.bounded_deque import BoundedDequeFullError

from lonelypss.ws.handlers.open.check_result import CheckResult
from lonelypss.ws.handlers.open.senders.send_any import send_any
from lonelypss.ws.state import (
    InternalLargeMessage,
    InternalMessageType,
    InternalSmallMessage,
    StateOpen,
    WaitingInternalMessageType,
    WaitingInternalSpooledLargeMessage,
)


async def check_my_receiver_queue(state: StateOpen) -> CheckResult:
    """Makes progress using the result of the read task, if possible. Raises
    an exception to indicate that we should begin the cleanup and shutdown
    process
    """
    if state.my_receiver.queue.empty():
        return CheckResult.CONTINUE

    result = state.my_receiver.queue.get_nowait()

    if state.send_task is None and not state.unsent_messages:
        state.send_task = asyncio.create_task(send_any(state, result))
        return CheckResult.RESTART

    try:
        state.unsent_messages.ensure_space_for(1)
    except BoundedDequeFullError:
        if result.type == InternalMessageType.LARGE:
            result.finished.set()
        raise

    if result.type != InternalMessageType.LARGE:
        state.unsent_messages.append(result)
        return CheckResult.RESTART

    spooled = _spool_large_message(state, result)
    state.unsent_messages.append(spooled)

    return CheckResult.RESTART


def _spool_large_message(
    state: StateOpen, message: InternalLargeMessage
) -> Union[InternalSmallMessage, WaitingInternalSpooledLargeMessage]:
    if message.length == 0:
        message.finished.set()
        return InternalSmallMessage(
            type=InternalMessageType.SMALL,
            topic=message.topic,
            data=b"",
            sha512=message.sha512,
        )

    if message.length < state.broadcaster_config.message_body_spool_size:
        try:
            message_body = message.stream.read(message.length)
        finally:
            message.finished.set()

        if len(message_body) != message.length:
            raise ValueError(
                f"expected {message.length} bytes, got {len(message_body)}"
            )

        return InternalSmallMessage(
            type=InternalMessageType.SMALL,
            topic=message.topic,
            data=message_body,
            sha512=message.sha512,
        )

    tmpfile = tempfile.TemporaryFile()
    try:
        remaining = message.length
        while remaining > 0:
            chunk = message.stream.read(min(io.DEFAULT_BUFFER_SIZE, remaining))
            if not chunk:
                raise ValueError(f"unexpected end of stream ({remaining} bytes left)")

            tmpfile.write(chunk)
            remaining -= len(chunk)

        if remaining < 0:
            raise ValueError(f"read too many bytes ({remaining} left)")

        tmpfile.seek(0)
        return WaitingInternalSpooledLargeMessage(
            type=WaitingInternalMessageType.SPOOLED_LARGE,
            stream=tmpfile,
            length=message.length,
            topic=message.topic,
            sha512=message.sha512,
        )
    except BaseException:
        tmpfile.close()
        raise
    finally:
        message.finished.set()



from typing import TYPE_CHECKING, Dict

from lonelypss.ws.handlers.accepting import handle_accepting
from lonelypss.ws.handlers.closing import handle_closing
from lonelypss.ws.handlers.open.handler import handle_open
from lonelypss.ws.handlers.protocol import StateHandler
from lonelypss.ws.handlers.waiting_configure import handle_waiting_configure
from lonelypss.ws.state import State, StateType

_handlers: Dict[StateType, StateHandler] = {
    StateType.ACCEPTING: handle_accepting,
    StateType.WAITING_CONFIGURE: handle_waiting_configure,
    StateType.OPEN: handle_open,
    StateType.CLOSING: handle_closing,
}


async def handle_any(state: State) -> State:
    """Handle any state by delegating to the appropriate handler. This will raise
    a KeyError for the CLOSED state, which does not need any further handling.

    Raises KeyError if no handler is found for the state type.
    """
    return await _handlers[state.type](state)


if TYPE_CHECKING:
    _: StateHandler = handle_any



import time
from typing import TYPE_CHECKING

from lonelypsp.auth.config import AuthResult
from lonelypsp.stateful.messages.continue_receive import S2B_ContinueReceive

from lonelypss.ws.handlers.open.processors.protocol import S2B_MessageProcessor
from lonelypss.ws.handlers.open.websocket_url import (
    make_for_receive_websocket_url_and_change_counter,
)
from lonelypss.ws.state import StateOpen


async def process_continue_receive(
    state: StateOpen, message: S2B_ContinueReceive
) -> None:
    expected_ack = state.expecting_acks.get_nowait()
    if expected_ack.type != message.type:
        raise Exception(
            f"unexpected continue receive (expecting a {expected_ack.type})"
        )
    if expected_ack.identifier != message.identifier:
        raise Exception(
            f"unexpected continue receive (expecting identifier {expected_ack.identifier!r}, got {message.identifier!r})"
        )
    if expected_ack.part_id != message.part_id:
        raise Exception(
            f"unexpected continue receive (expecting part_id {expected_ack.part_id}, got {message.part_id})"
        )

    receive_url = make_for_receive_websocket_url_and_change_counter(state)
    auth_result = await state.broadcaster_config.is_stateful_continue_receive_allowed(
        url=receive_url,
        message=message,
        now=time.time(),
    )
    if auth_result != AuthResult.OK:
        raise Exception(f"continue receive auth: {auth_result}")


if TYPE_CHECKING:
    _: S2B_MessageProcessor[S2B_ContinueReceive] = process_continue_receive



import time
from typing import TYPE_CHECKING

from lonelypsp.auth.config import AuthResult
from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType
from lonelypsp.stateful.messages.confirm_unsubscribe import (
    B2S_ConfirmUnsubscribeExact,
    serialize_b2s_confirm_unsubscribe_exact,
)
from lonelypsp.stateful.messages.unsubscribe import S2B_UnsubscribeExact

from lonelypss.ws.handlers.open.errors import AuthRejectedException
from lonelypss.ws.handlers.open.processors.protocol import S2B_MessageProcessor
from lonelypss.ws.handlers.open.send_simple_asap import send_simple_asap
from lonelypss.ws.handlers.open.websocket_url import (
    make_for_receive_websocket_url_and_change_counter,
)
from lonelypss.ws.state import StateOpen


async def process_unsubscribe_exact(
    state: StateOpen, message: S2B_UnsubscribeExact
) -> None:
    """Processes a request by the subscriber to unsubscribe from a specific topic,
    no longer receiving notifications within this websocket
    """
    url = make_for_receive_websocket_url_and_change_counter(state)
    auth_at = time.time()
    auth_result = await state.broadcaster_config.is_subscribe_exact_allowed(
        tracing=message.tracing,
        url=url,
        recovery=None,
        exact=message.topic,
        now=auth_at,
        authorization=message.authorization,
    )
    if auth_result != AuthResult.OK:
        raise AuthRejectedException(f"unsubscribe exact: {auth_result}")

    if message.topic not in state.my_receiver.exact_subscriptions:
        raise Exception("not subscribed to exact topic")

    resp_tracing = b""  # TODO: tracing
    # TODO: not safe to use a different url here I believe, because we would need to
    # ensure nothing else gets queued to be sent between now and the send actually
    # being queued, which would require being in a send_task, not a process_task
    resp_authorization = (
        await state.broadcaster_config.authorize_confirm_subscribe_exact(
            tracing=resp_tracing,
            url=url,
            recovery=None,
            exact=message.topic,
            now=time.time(),
        )
    )

    state.my_receiver.exact_subscriptions.remove(message.topic)
    await state.internal_receiver.decrement_exact(message.topic)
    send_simple_asap(
        state,
        serialize_b2s_confirm_unsubscribe_exact(
            B2S_ConfirmUnsubscribeExact(
                type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_UNSUBSCRIBE_EXACT,
                topic=message.topic,
                authorization=resp_authorization,
                tracing=resp_tracing,
            ),
            minimal_headers=state.broadcaster_config.websocket_minimal_headers,
        ),
    )


if TYPE_CHECKING:
    _: S2B_MessageProcessor[S2B_UnsubscribeExact] = process_unsubscribe_exact



from lonelypss.config.config import Config


async def setup_config(config: Config) -> None:
    """Convenience function to setup the configuration (similiar idea to aenter)"""
    await config.setup_to_broadcaster_auth()
    try:
        await config.setup_to_subscriber_auth()
        try:
            await config.setup_db()
            try:
                await config.setup_http_notify_client_session(config)
                try:
                    await config.setup_tracing()
                except BaseException:
                    await config.teardown_http_notify_client_session()
                    raise
            except BaseException:
                await config.teardown_db()
                raise
        except BaseException:
            await config.teardown_to_subscriber_auth()
            raise
    except BaseException:
        await config.teardown_to_broadcaster_auth()
        raise


async def teardown_config(config: Config) -> None:
    """Convenience function to teardown the configuration (similiar idea to aenter)"""
    try:
        await config.teardown_tracing()
    finally:
        try:
            await config.teardown_http_notify_client_session()
        finally:
            try:
                await config.teardown_db()
            finally:
                try:
                    await config.teardown_to_subscriber_auth()
                finally:
                    await config.teardown_to_broadcaster_auth()



import time
from typing import TYPE_CHECKING

from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType
from lonelypsp.stateful.messages.disable_zstd_custom import (
    B2S_DisableZstdCustom,
    serialize_b2s_disable_zstd_custom,
)

from lonelypss.ws.handlers.open.senders.protocol import Sender
from lonelypss.ws.handlers.open.websocket_url import (
    make_for_send_websocket_url_and_change_counter,
)
from lonelypss.ws.state import SimplePendingSendDisableZstdCustom, StateOpen


async def send_disable_zstd_custom(
    state: StateOpen, message: SimplePendingSendDisableZstdCustom
) -> None:
    """Sends the message to the ASGI server for forwarding"""
    # we need this in send lock because theres an event loop between url and sending
    # the message, and during that time it's important nothing else produces a send
    # url
    url = make_for_send_websocket_url_and_change_counter(state)
    tracing = b""  # TODO: tracing
    authorization = (
        await state.broadcaster_config.authorize_stateful_disable_zstd_custom(
            tracing=tracing,
            compressor_identifier=message.identifier,
            url=url,
            now=time.time(),
        )
    )
    data = serialize_b2s_disable_zstd_custom(
        B2S_DisableZstdCustom(
            type=BroadcasterToSubscriberStatefulMessageType.DISABLE_ZSTD_CUSTOM,
            identifier=message.identifier,
            authorization=authorization,
            tracing=tracing,
        ),
        minimal_headers=state.broadcaster_config.websocket_minimal_headers,
    )
    await state.websocket.send_bytes(data)


if TYPE_CHECKING:
    _: Sender[SimplePendingSendDisableZstdCustom] = send_disable_zstd_custom



import asyncio
import os
import tempfile
import time
from types import TracebackType
from typing import TYPE_CHECKING, List, Optional, Protocol, Tuple, Type

from lonelypss.util.sync_io import read_exact
from lonelypss.ws.handlers.open.send_simple_asap import send_asap
from lonelypss.ws.state import (
    Compressor,
    CompressorPreparing,
    CompressorReady,
    CompressorState,
    CompressorTrainingDataCollector,
    CompressorTrainingInfoBeforeHighWatermark,
    CompressorTrainingInfoType,
    CompressorTrainingInfoWaitingToRefresh,
    SimplePendingSendDisableZstdCustom,
    SimplePendingSendType,
    StateOpen,
)

try:
    import zstandard
except ImportError:
    ...


class _ConfigCompressorGenerator(Protocol):
    """
    type that describes
    state.broadcaster_config.train_compression_dict_low_watermark and
    state.broadcaster_config.train_compression_dict_high_watermark
    """

    async def __call__(
        self, /, samples: List[bytes]
    ) -> "Tuple[zstandard.ZstdCompressionDict, int]": ...


async def make_ready_compressor_from_collector_and_generator(
    state: StateOpen,
    collector: CompressorTrainingDataCollector,
    generator: _ConfigCompressorGenerator,
    identifier: int,
    close_fd: bool,
) -> CompressorReady:
    """Generates a new ready compressor using the data in the given collector
    and the given compression dictionary generator from samples. This will
    load samples from the collector, allowing for new data to be written to the
    collector in the meantime and regularly yielding to the event loop

    Args:
        state: the state object
        collector: the collector with the training data
        generator: the generator that will create the compression dictionary
        identifier: the identifier for the compressor we are making
        close_fd: true to close the collectors tmpfile when done, regardless of
            success, false not to
    """

    our_event = asyncio.Event()
    try:
        # while we are running new data may be written to the collector;
        # in order to ensure that we don't get continuously stuck, we will
        # only wait until the data that is already pending in the collector
        # is ready while also preventing the tmpfile from being closed before
        # we've read that data

        usable_num_messages = collector.messages
        writing_to_usable = set(
            asyncio.create_task(event.wait()) for event in collector.pending
        )
        collector.pending.add(our_event)

        await asyncio.wait(writing_to_usable, return_when=asyncio.ALL_COMPLETED)
        for wait in writing_to_usable:
            wait.result()

        samples: List[bytes] = []
        pos = 0
        while len(samples) < usable_num_messages:
            collector.tmpfile.seek(pos, os.SEEK_SET)
            sample_len_bytes = read_exact(collector.tmpfile, 4)
            sample_len = int.from_bytes(sample_len_bytes, "big")
            sample = read_exact(collector.tmpfile, sample_len)
            samples.append(sample)
            pos += 4 + sample_len
            await asyncio.sleep(0)

        our_event.set()
        collector.pending.discard(our_event)

        zdict, level = await generator(samples)
        return CompressorReady(
            type=CompressorState.READY,
            identifier=identifier,
            level=level,
            min_size=state.broadcaster_config.compression_min_size,
            max_size=state.broadcaster_config.compression_trained_max_size,
            data=zdict,
            compressors=list(),
            decompressors=list(),
        )
    finally:
        our_event.set()
        collector.pending.discard(our_event)

        if close_fd:
            collector.tmpfile.close()


def make_preparing_compressor_from_collector_and_generator(
    state: StateOpen,
    collector: CompressorTrainingDataCollector,
    generator: _ConfigCompressorGenerator,
    identifier: int,
    close_fd: bool,
) -> CompressorPreparing:
    """Creates a preparing compressor that will use the data within the given
    collector to generate a new compression dictionary with the given generator,
    then use that to make a compressor and decompressor

    Args:
        state: the state object
        collector: the collector with the training data
        generator: the generator that will create the compression dictionary
        identifier: the identifier for the compressor
        close_fd: true to close the collectors tmpfile when done, regardless of
            success, false not to
    """
    return CompressorPreparing(
        type=CompressorState.PREPARING,
        identifier=identifier,
        task=asyncio.create_task(
            make_ready_compressor_from_collector_and_generator(
                state, collector, generator, identifier, close_fd
            )
        ),
    )


def rotate_in_compressor(state: StateOpen, compressor: Compressor) -> None:
    """Removes the a non-preset compressor in the state, if there are
    at least 2, and adds the given compressor to the state

    Assuming this is the only way non-preset compressors are added to
    the state, this will always remove the oldest non-preset compressor
    """

    first = True
    for candidate_idx in range(len(state.compressors) - 1, -1, -1):
        candidate = state.compressors[candidate_idx]
        if candidate.identifier < 65536:
            continue

        if first:
            first = False
            continue

        state.compressors.pop(candidate_idx)
        if candidate.type == CompressorState.PREPARING:
            candidate.task.cancel()

        send_asap(
            state,
            SimplePendingSendDisableZstdCustom(
                type=SimplePendingSendType.DISABLE_ZSTD_CUSTOM,
                identifier=candidate.identifier,
            ),
        )
        break

    state.compressors.append(compressor)


def maybe_advance_compressor_training_info(state: StateOpen) -> None:
    """If we are collecting data for training a compression dictionary and we
    have enough training data, rotates the new compressor in and advances
    the compressor training info
    """
    if state.compressor_training_info is None:
        return

    if (
        state.compressor_training_info.type
        == CompressorTrainingInfoType.WAITING_TO_REFRESH
    ):
        now = time.time()
        if (
            state.compressor_training_info.last_built
            + state.broadcaster_config.compression_retrain_interval_seconds
            > now
        ):
            return

        tmpfile = tempfile.SpooledTemporaryFile(
            max_size=state.broadcaster_config.message_body_spool_size
        )
        try:
            state.compressor_training_info = CompressorTrainingInfoBeforeHighWatermark(
                type=CompressorTrainingInfoType.BEFORE_HIGH_WATERMARK,
                collector=CompressorTrainingDataCollector(
                    messages=0, length=0, tmpfile=tmpfile, pending=set()
                ),
                compressor_id=state.compressor_training_info.compressor_id,
            )
            return
        except BaseException:
            tmpfile.close()
            raise

    if (
        state.compressor_training_info.type
        == CompressorTrainingInfoType.BEFORE_LOW_WATERMARK
    ):
        if (
            state.compressor_training_info.collector.length
            < state.broadcaster_config.compression_training_low_watermark
        ):
            return

        if (
            state.compressor_training_info.collector.length
            < state.broadcaster_config.compression_training_high_watermark
        ):
            rotate_in_compressor(
                state,
                make_preparing_compressor_from_collector_and_generator(
                    state,
                    state.compressor_training_info.collector,
                    state.broadcaster_config.train_compression_dict_low_watermark,
                    identifier=state.compressor_training_info.compressor_id,
                    close_fd=False,
                ),
            )
            state.compressor_training_info = CompressorTrainingInfoBeforeHighWatermark(
                type=CompressorTrainingInfoType.BEFORE_HIGH_WATERMARK,
                collector=state.compressor_training_info.collector,
                compressor_id=state.compressor_training_info.compressor_id + 1,
            )
            return

        state.compressor_training_info = CompressorTrainingInfoBeforeHighWatermark(
            type=CompressorTrainingInfoType.BEFORE_HIGH_WATERMARK,
            collector=state.compressor_training_info.collector,
            compressor_id=state.compressor_training_info.compressor_id,
        )

    if (
        state.compressor_training_info.collector.length
        < state.broadcaster_config.compression_training_high_watermark
    ):
        return

    rotate_in_compressor(
        state,
        make_preparing_compressor_from_collector_and_generator(
            state,
            state.compressor_training_info.collector,
            state.broadcaster_config.train_compression_dict_high_watermark,
            identifier=state.compressor_training_info.compressor_id,
            close_fd=True,
        ),
    )
    state.compressor_training_info = CompressorTrainingInfoWaitingToRefresh(
        type=CompressorTrainingInfoType.WAITING_TO_REFRESH,
        last_built=time.time(),
        compressor_id=state.compressor_training_info.compressor_id + 1,
    )


def maybe_store_small_message_for_training(state: StateOpen, data: bytes) -> None:
    """If we are collecting data for training a compression dictionary, store
    the given data from memory synchronously. These can be weaved in at any time
    even while large messages are being written to the collector over multiple
    event loops.
    """
    if state.compressor_training_info is None:
        return

    if len(data) < state.broadcaster_config.compression_min_size:
        return

    if len(data) >= state.broadcaster_config.compression_trained_max_size:
        return

    if (
        state.compressor_training_info.type
        == CompressorTrainingInfoType.WAITING_TO_REFRESH
    ):
        maybe_advance_compressor_training_info(state)
        if (
            state.compressor_training_info.type
            == CompressorTrainingInfoType.WAITING_TO_REFRESH
        ):
            return

    collector = state.compressor_training_info.collector

    try:
        collector.tmpfile.seek(0, os.SEEK_END)
        collector.tmpfile.write(len(data).to_bytes(4, "big"))
        collector.tmpfile.write(data)
    except BaseException:
        collector.tmpfile.close()
        raise


class CompressorLargeMessageWriter(Protocol):
    """Convenience protocol for writing large messages to the training
    collector, generated by maybe_write_large_message_for_training

    For convenience this acts as a synchronous context manager where enter
    does nothing and exit calls raise_if_not_done
    """

    @property
    def length(self) -> int:
        """the length of the message we are writing"""

    @property
    def remaining(self) -> int:
        """how many more bytes we expect to receive"""

    @property
    def is_void(self) -> bool:
        """true if this is a void writer, false if you need to write"""

    def skip_void(self) -> None:
        """Raises an error unless this is a void writer, in which case this
        sets remaining to zero
        """

    def raise_if_not_done(self) -> None:
        """cleans up and raises an error if remaining is not zero"""

    def write_chunk(self, data: bytes) -> None:
        """Writes a chunk of the message to the collector. If this is
        the last chunk, cleans up resources. To check if this was the
        last chunk, remaining will be zero
        """

    def __enter__(self) -> "CompressorLargeMessageWriter":
        """no-op"""

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """calls raise_if_not_done"""


class VoidingCompressorLargeMessageWriter:
    def __init__(self, length: int) -> None:
        self.length = length
        self.remaining = length
        self.is_void = True

    def skip_void(self) -> None:
        self.remaining = 0

    def raise_if_not_done(self) -> None:
        if self.remaining != 0:
            raise ValueError("not done writing")

    def write_chunk(self, data: bytes) -> None:
        if self.remaining < len(data):
            raise ValueError("too much data")
        self.remaining -= len(data)

    def __enter__(self) -> "CompressorLargeMessageWriter":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.raise_if_not_done()


if TYPE_CHECKING:
    _: Type[CompressorLargeMessageWriter] = VoidingCompressorLargeMessageWriter


class _RealCompressorLargeMessageWriter:
    def __init__(self, collector: CompressorTrainingDataCollector, length: int) -> None:
        self._collector = collector
        self._event = asyncio.Event()
        self._collector.pending.add(self._event)
        self._pos = -1
        self.length = length
        self.remaining = length
        self.is_void = False

        try:
            self._pos = collector.tmpfile.seek(0, os.SEEK_END)
            collector.tmpfile.write(length.to_bytes(4, "big"))
        except BaseException:
            self._cleanup()
            collector.tmpfile.close()
            raise

        self._pos += 4

        if length == 0:
            self._cleanup()

    def skip_void(self) -> None:
        raise Exception("not a void writer")

    def raise_if_not_done(self) -> None:
        if self.remaining != 0:
            self._cleanup()
            raise ValueError("not done writing")

    def write_chunk(self, data: bytes) -> None:
        if self.remaining < len(data):
            self._cleanup()
            raise ValueError("too much data")

        self.remaining -= len(data)
        try:
            self._collector.tmpfile.seek(self._pos, os.SEEK_SET)
            self._collector.tmpfile.write(data)
        except BaseException:
            self._cleanup()
            self._collector.tmpfile.close()
            raise

        self._pos += len(data)
        if self.remaining == 0:
            self._cleanup()

    def _cleanup(self) -> None:
        self._event.set()
        self._collector.pending.discard(self._event)

    def __enter__(self) -> "CompressorLargeMessageWriter":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.raise_if_not_done()


if TYPE_CHECKING:
    __: Type[CompressorLargeMessageWriter] = _RealCompressorLargeMessageWriter


def maybe_write_large_message_for_training(
    state: StateOpen, length: int, never_store: bool = False
) -> CompressorLargeMessageWriter:
    """If we need to collect a message of the given length in the training
    collector, returns an object that will write a large message to the
    collector. Otherwise, return an object with the same protocol that will just
    verify it is being used correctly.

    Multiple of these writers can be used at the same time on the main asyncio
    thread, even while training is occuring (on the main asyncio thread), and it
    may be weaved with small messages being written to the collector (again, on
    the main asyncio thread)

    If `never_store` is set this will always return a void writer, which can be
    convenient when using this as a context manager
    """

    if (
        never_store
        or state.compressor_training_info is None
        or length < state.broadcaster_config.compression_min_size
        or length >= state.broadcaster_config.compression_trained_max_size
    ):
        return VoidingCompressorLargeMessageWriter(length)

    if (
        state.compressor_training_info.type
        == CompressorTrainingInfoType.WAITING_TO_REFRESH
    ):
        maybe_advance_compressor_training_info(state)
        if (
            state.compressor_training_info.type
            == CompressorTrainingInfoType.WAITING_TO_REFRESH
        ):
            return VoidingCompressorLargeMessageWriter(length)

    return _RealCompressorLargeMessageWriter(
        state.compressor_training_info.collector, length
    )



import time
from typing import Annotated, Optional

from fastapi import APIRouter, Header, Request, Response
from lonelypsp.auth.config import AuthResult
from lonelypsp.compat import assert_never
from lonelypsp.stateless.constants import BroadcasterToSubscriberStatelessMessageType
from lonelypsp.sync_io import PreallocatedBytesIO

from lonelypss.middleware.config import get_config_from_request
from lonelypss.util.async_io import async_read_exact
from lonelypss.util.request_body_io import AsyncIterableAIO

router = APIRouter()


@router.post(
    "/v1/unsubscribe/exact",
    status_code=200,
    responses={
        "400": {"description": "The body was not formatted correctly"},
        "401": {"description": "Authorization header is required but not provided"},
        "403": {"description": "Authorization header is provided but invalid"},
        "409": {"description": "The subscription does not exist"},
        "500": {"description": "Unexpected error occurred"},
        "503": {"description": "Service is unavailable, try again soon"},
    },
)
async def unsubscribe_exact(
    request: Request, authorization: Annotated[Optional[str], Header()] = None
) -> Response:
    """Unsubscribes the given URL from the given topic. The body should be
    formatted as the following sequence:

    - 2 bytes (N): length of the url that the subscriber can be reached at, big-endian, unsigned
    - N bytes: the url that the subscriber can be reached at, utf-8 encoded
    - 2 bytes (M): the length of the topic, big-endian, unsigned
    - M bytes: the topic
    - 2 bytes (T): length of the tracing data, big-endian, unsigned
    - T bytes: the tracing data

    The response has an arbitrary body (generally empty) and one of the
    following status codes:

    - 200 Okay: the subscription was removed
    - 400 Bad Request: the body was not formatted correctly
    - 401 Unauthorized: authorization is required but not provided
    - 403 Forbidden: authorization is provided but invalid
    - 409 Conflict: the subscription already exists
    - 500 Internal Server Error: unexpected error occurred
    - 503 Service Unavailable: servce (generally, database) is unavailable

    ### response body for 200 Okay
    - `BroadcasterToSubscriberStatelessMessageType.RESPONSE_GENERIC`
    """
    config = get_config_from_request(request)

    try:
        stream = request.stream()
        try:
            body = AsyncIterableAIO(stream.__aiter__())

            url_length_bytes = await async_read_exact(body, 2)
            url_length = int.from_bytes(url_length_bytes, "big")
            url_bytes = await async_read_exact(body, url_length)
            url = url_bytes.decode("utf-8")

            topic_length_bytes = await async_read_exact(body, 2)
            topic_length = int.from_bytes(topic_length_bytes, "big")
            topic = await async_read_exact(body, topic_length)

            tracing_length_bytes = await async_read_exact(body, 2)
            tracing_length = int.from_bytes(tracing_length_bytes, "big")
            tracing = await async_read_exact(body, tracing_length)
        finally:
            await stream.aclose()
    except ValueError:
        return Response(status_code=400)

    auth_at = time.time()
    auth_result = await config.is_subscribe_exact_allowed(
        tracing=tracing,
        url=url,
        recovery=None,
        exact=topic,
        now=auth_at,
        authorization=authorization,
    )

    if auth_result == AuthResult.UNAUTHORIZED:
        return Response(status_code=401)
    elif auth_result == AuthResult.FORBIDDEN:
        return Response(status_code=403)
    elif auth_result == AuthResult.UNAVAILABLE:
        return Response(status_code=503)
    elif auth_result != AuthResult.OK:
        assert_never(auth_result)

    db_result = await config.unsubscribe_exact(url=url, exact=topic)

    if db_result == "not_found":
        return Response(status_code=409)
    elif db_result == "unavailable":
        return Response(status_code=503)
    elif db_result != "success":
        assert_never(db_result)

    resp_tracing = b""  # TODO: tracing
    resp_authorization = await config.authorize_confirm_subscribe_exact(
        tracing=resp_tracing,
        url=url,
        recovery=None,
        exact=topic,
        now=time.time(),
    )
    resp_authorization_bytes = (
        b"" if resp_authorization is None else resp_authorization.encode("utf-8")
    )

    resp_body = PreallocatedBytesIO(
        2 + 2 + len(resp_authorization_bytes) + 2 + len(resp_tracing)
    )
    resp_body.write(
        int(BroadcasterToSubscriberStatelessMessageType.RESPONSE_GENERIC).to_bytes(
            2, "big"
        )
    )
    resp_body.write(len(resp_authorization_bytes).to_bytes(2, "big"))
    resp_body.write(resp_authorization_bytes)
    resp_body.write(len(resp_tracing).to_bytes(2, "big"))
    resp_body.write(resp_tracing)

    return Response(
        status_code=200,
        headers={"Content-Type": "application/octet-stream"},
        content=memoryview(resp_body.buffer),
    )



import asyncio
from typing import List, Set

from lonelypss.ws.handlers.open.check_result import CheckResult
from lonelypss.ws.handlers.open.errors import combine_multiple_exceptions
from lonelypss.ws.state import StateOpen


async def check_background_tasks(state: StateOpen) -> CheckResult:
    """Checks to see if any of the background tasks have finished so we can
    check for errors and release the references to them.

    Raises an exception to indicate we should move to the cleanup and disconnect
    process
    """

    for task in state.backgrounded:
        if task.done():
            break
    else:
        return CheckResult.CONTINUE

    done: Set[asyncio.Task[None]] = set()
    pending: Set[asyncio.Task[None]] = set()

    for task in state.backgrounded:
        if task.done():
            done.add(task)
        else:
            pending.add(task)

    state.backgrounded = pending
    excs: List[BaseException] = []
    for finished in done:
        try:
            finished.result()
        except BaseException as e:
            excs.append(e)

    if excs:
        raise combine_multiple_exceptions("error in background task", excs)

    return CheckResult.RESTART



import asyncio
from typing import TYPE_CHECKING

from lonelypss.ws.state import (
    SimplePendingSend,
    SimplePendingSendPreFormatted,
    SimplePendingSendType,
    StateOpen,
)

if TYPE_CHECKING:
    from lonelypss.ws.handlers.open.senders.send_any import send_any
else:
    send_any = None


def send_asap(state: StateOpen, pending: SimplePendingSend) -> None:
    """Queues a simple message to be sent to the subscriber as soon as possible,
    where that message has already been described via one of the SimplePendingSend
    options
    """
    global send_any
    if send_any is None:
        from lonelypss.ws.handlers.open.senders.send_any import send_any

    if state.send_task is None:
        state.send_task = asyncio.create_task(send_any(state, pending))
        return

    state.unsent_messages.append(pending)


def send_simple_asap(state: StateOpen, data: bytes) -> None:
    """Queues a simple message to be sent to the subscriber as soon as possible"""

    if state.send_task is None:
        state.send_task = asyncio.create_task(state.websocket.send_bytes(data))
        return

    state.unsent_messages.append(
        SimplePendingSendPreFormatted(
            type=SimplePendingSendType.PRE_FORMATTED, data=data
        )
    )



import base64
import hashlib
import io
import logging
import secrets
import tempfile
import time
from dataclasses import dataclass
from enum import Enum, auto
from typing import Annotated, Dict, Generic, Literal, Optional, TypeVar, Union

import aiohttp
from fastapi import APIRouter, Header, Request, Response
from lonelypsp.auth.config import AuthResult
from lonelypsp.compat import assert_never, fast_dataclass
from lonelypsp.stateless.constants import (
    BroadcasterToSubscriberStatelessMessageType,
    SubscriberToBroadcasterStatelessMessageType,
)
from lonelypsp.sync_io import PreallocatedBytesIO
from lonelypsp.tracing.shared.handle_trusted_notify import (
    HandledTrustedNotify,
    HandledTrustedNotifyHandleMissedStart,
)

from lonelypss.config.config import (
    Config,
    MissedInfo,
    SubscriberInfoExact,
    SubscriberInfoGlob,
    SubscriberInfoType,
)
from lonelypss.middleware.config import get_config_from_request
from lonelypss.util.async_io import AsyncReadableBytesIO, async_read_exact
from lonelypss.util.close_guarded_io import CloseGuardedIO
from lonelypss.util.sync_io import SyncIOBaseLikeIO, read_exact

router = APIRouter()


@router.post(
    "/v1/notify",
    status_code=200,
    responses={
        "400": {"description": "The body was not formatted correctly"},
        "401": {"description": "Authorization header is required but not provided"},
        "403": {"description": "Authorization header is provided but invalid"},
        "500": {"description": "Unexpected error occurred"},
        "503": {"description": "Service is unavailable, try again soon"},
    },
)
async def notify(
    request: Request, authorization: Annotated[Optional[str], Header()] = None
) -> Response:
    """Sends the given message to subscribers for the given topic. The body should be
    formatted as the following sequence:

    - 2 bytes (N): length of the topic, big-endian, unsigned
    - N bytes: the topic. if utf-8 decodable then we will attempt to match glob
      patterns, otherwise, only goes to exact subscriptions
    - 64 bytes: sha-512 hash of the message, will be rechecked
    - 2 bytes (T): length of the tracing data, big-endian, unsigned
    - T bytes: the tracing data
    - 1 bytes (I): length of the identifier, big-endian, unsigned
    - I bytes: the identifier
    - 8 bytes (M): length of the message, big-endian, unsigned
    - M bytes: the message

    The response has one of the following status codes, where the body is arbitrary
    unless otherwise specified.

    - 200 Okay: subscribers were notified. response is as follows:
        - 2 bytes (type): int(RESPONSE_NOTIFY), big endian, unsigned
        - 2 bytes (A): big-endian, unsigned, the length of the authorization. broadcaster
            side authorization is always used because hmac over http is supported
        - A bytes: the authorization
        - 2 bytes (T): big-endian, unsigned, the length of tracing data
        - T bytes: the tracing data
        - 2 bytes: big-endian, unsigned, the number of subscribers notified
        - 1 byte (I): length of the identifier, big-endian, unsigned
        - I bytes: the identifier
    - 400 Bad Request: the body was not formatted correctly
    - 401 Unauthorized: authorization is required but not provided
    - 403 Forbidden: authorization is provided but invalid
    - 500 Internal Server Error: unexpected error occurred
    - 503 Service Unavailable: servce (generally, database) is unavailable
    """
    config = get_config_from_request(request)

    with config.tracer.stateless.receive_notify(None) as initial_trace:
        on_auth_result_trace = initial_trace.on_received()
        del initial_trace

        with tempfile.SpooledTemporaryFile(
            max_size=config.message_body_spool_size, mode="w+b"
        ) as request_body:
            read_length = 0
            saw_end = False

            stream_iter = request.stream().__aiter__()
            while True:
                try:
                    chunk = await stream_iter.__anext__()
                except StopAsyncIteration:
                    saw_end = True
                    break

                request_body.write(chunk)
                read_length += len(chunk)
                if read_length >= 2 + 65535 + 64 + 2 + 65535 + 1 + 255 + 8:
                    break

            request_body.seek(0)
            try:
                topic_length = int.from_bytes(read_exact(request_body, 2), "big")
                topic = read_exact(request_body, topic_length)
                message_hash = read_exact(request_body, 64)
                tracing_length = int.from_bytes(read_exact(request_body, 2), "big")
                tracing = read_exact(request_body, tracing_length)
                identifier_length = int.from_bytes(read_exact(request_body, 1), "big")
                identifier = read_exact(request_body, identifier_length)
                message_length = int.from_bytes(read_exact(request_body, 8), "big")
            except ValueError:
                on_auth_result_trace.on_bad_request()
                return Response(status_code=400)
            header_length = request_body.tell()

            auth_at = time.time()
            auth_result = await config.is_notify_allowed(
                tracing=tracing,
                topic=topic,
                identifier=identifier,
                message_sha512=message_hash,
                now=auth_at,
                authorization=authorization,
            )

            if auth_result == AuthResult.UNAUTHORIZED:
                on_auth_result_trace.on_bad_auth_result(result=auth_result)
                return Response(status_code=401)
            elif auth_result == AuthResult.FORBIDDEN:
                on_auth_result_trace.on_bad_auth_result(result=auth_result)
                return Response(status_code=403)
            elif auth_result == AuthResult.UNAVAILABLE:
                on_auth_result_trace.on_bad_auth_result(result=auth_result)
                return Response(status_code=503)
            elif auth_result != AuthResult.OK:
                assert_never(auth_result)

            hasher = hashlib.sha512()
            on_auth_verified_trace = on_auth_result_trace.on_auth_tentatively_accepted()
            del on_auth_result_trace

            while True:
                chunk = request_body.read(io.DEFAULT_BUFFER_SIZE)
                if not chunk:
                    break
                hasher.update(chunk)

            if not saw_end:
                while True:
                    try:
                        chunk = await stream_iter.__anext__()
                    except StopAsyncIteration:
                        saw_end = True
                        break

                    hasher.update(chunk)
                    request_body.write(chunk)
                    read_length += len(chunk)

                    if read_length > header_length + message_length:
                        on_auth_verified_trace.on_auth_mismatch()
                        return Response(status_code=400)

            if read_length != header_length + message_length:
                on_auth_verified_trace.on_auth_mismatch()
                return Response(status_code=400)

            actual_hash = hasher.digest()
            if actual_hash != message_hash:
                on_auth_verified_trace.on_auth_mismatch()
                return Response(status_code=400)

            handle_trusted_notify_trace = on_auth_verified_trace.on_auth_accepted(
                topic=topic,
                length=message_length,
                identifier=identifier,
                tracing=tracing,
            )
            del on_auth_verified_trace

            request_body.seek(2 + topic_length + 64 + 8)
            notify_result = await handle_trusted_notify(
                topic,
                request_body,
                config=config,
                session=config.http_notify_client_session,
                content_length=message_length,
                sha512=actual_hash,
                tracer=handle_trusted_notify_trace,
            )
            del handle_trusted_notify_trace

            if notify_result.type == TrustedNotifyResultType.UNAVAILABLE:
                return Response(status_code=503)

            resp_tracing = notify_result.tracer.on_sending_response()
            resp_authorization = await config.authorize_confirm_notify(
                tracing=resp_tracing,
                identifier=identifier,
                subscribers=notify_result.succeeded,
                topic=topic,
                message_sha512=message_hash,
                now=time.time(),
            )
            resp_authorization_bytes = (
                b""
                if resp_authorization is None
                else resp_authorization.encode("utf-8")
            )
            resp = PreallocatedBytesIO(
                2
                + 2
                + len(resp_authorization_bytes)
                + 2
                + len(resp_tracing)
                + 4
                + 1
                + len(identifier)
            )
            resp.write(
                int(
                    BroadcasterToSubscriberStatelessMessageType.RESPONSE_NOTIFY
                ).to_bytes(2, "big")
            )
            resp.write(len(resp_authorization_bytes).to_bytes(2, "big"))
            resp.write(resp_authorization_bytes)
            resp.write(len(resp_tracing).to_bytes(2, "big"))
            resp.write(resp_tracing)
            resp.write(notify_result.succeeded.to_bytes(4, "big"))
            resp.write(len(identifier).to_bytes(1, "big"))
            resp.write(identifier)

            return Response(
                status_code=200,
                content=memoryview(resp.buffer),
                headers={
                    "Content-Type": "application/octet-stream",
                },
            )


class TrustedNotifyResultType(Enum):
    UNAVAILABLE = auto()
    """We had trouble accessing the data store and may not have attempted all
    subscribers
    """
    OK = auto()
    """We at least attempted all subscribers"""


T = TypeVar("T")


@dataclass
class TrustedNotifyResultOK(Generic[T]):
    type: Literal[TrustedNotifyResultType.OK]
    """discriminator type"""
    succeeded: int
    """The number of subscribers we reached"""
    failed: int
    """The number of subscribers we could not reach"""
    tracer: T
    """The new tracer"""


@dataclass
class TrustedNotifyResultUnavailable:
    type: Literal[TrustedNotifyResultType.UNAVAILABLE]
    """discriminator type"""
    partial_succeeded: int
    """The number of subscribers we reached"""
    partial_failed: int
    """The number of subscribers we could not reach"""


async def _handle_missed(
    config: Config,
    topic: bytes,
    subscriber: Union[SubscriberInfoExact, SubscriberInfoGlob],
    tracer: HandledTrustedNotifyHandleMissedStart[T],
) -> T:
    on_done_tracer = tracer.on_handle_missed_start()
    del tracer

    if subscriber.recovery is None:
        return on_done_tracer.on_handle_missed_skipped(
            recovery=subscriber.recovery,
            next_retry_at=None,
        )

    next_retry_at = await config.get_delay_for_next_missed_retry(
        receive_url=subscriber.url,
        missed_url=subscriber.recovery,
        topic=topic,
        attempts=0,
    )
    if next_retry_at is None:
        return on_done_tracer.on_handle_missed_skipped(
            recovery=subscriber.recovery, next_retry_at=None
        )

    db_result = await config.upsert_missed(
        info=MissedInfo(
            topic=topic,
            attempts=0,
            next_retry_at=next_retry_at,
            subscriber_info=subscriber,
        )
    )

    if db_result == "unavailable":
        return on_done_tracer.on_handle_missed_unavailable(
            recovery=subscriber.recovery,
            next_retry_at=next_retry_at,
        )
    if db_result != "success":
        assert_never(db_result)

    return on_done_tracer.on_handle_missed_success(
        recovery=subscriber.recovery,
        next_retry_at=next_retry_at,
    )


async def handle_trusted_notify(
    topic: bytes,
    data: SyncIOBaseLikeIO,
    /,
    *,
    config: Config,
    session: aiohttp.ClientSession,
    content_length: int,
    sha512: bytes,
    tracer: HandledTrustedNotify[T],
) -> Union[TrustedNotifyResultOK[T], TrustedNotifyResultUnavailable]:
    """Notifies subscribers to the given topic with the given data.

    Args:
        topic (bytes): the topic the message was sent to
        data (file-like, readable, bytes): the message that was sent
        config (Config): the broadcaster configuration to use
        session (aiohttp.ClientSession): the aiohttp client session to
            send requests to clients in
        content_length (int): the length of the message in bytes. The stream
            MUST not have more than this amount of data in it, as we will read
            past this length if it does
        sha512 (bytes): the sha512 hash of the content (64 bytes)
        tracer (HandledTrustedNotify[T]): the tracer for debugging. should be
            considered unusable after this call (if OK, the return value has the new
            tracer)
    """
    succeeded = 0
    failed = 0
    headers: Dict[str, str] = {
        "Content-Type": "application/octet-stream",
        "Content-Length": str(content_length),
        "Repr-Digest": f"sha-512={base64.b64encode(sha512).decode('ascii')}",
        "X-Topic": base64.b64encode(topic).decode("ascii"),
    }

    message_starts_at = data.tell()
    guarded_request_body = CloseGuardedIO(data)

    async for subscriber in config.get_subscribers(topic=topic):
        if subscriber.type == SubscriberInfoType.UNAVAILABLE:
            tracer.on_unavailable()
            return TrustedNotifyResultUnavailable(
                type=TrustedNotifyResultType.UNAVAILABLE,
                partial_succeeded=succeeded,
                partial_failed=failed,
            )

        if subscriber.type == SubscriberInfoType.EXACT:
            sending_receive_tracer = tracer.on_exact_subscriber_found(
                url=subscriber.url
            )
            del tracer
        elif subscriber.type == SubscriberInfoType.GLOB:
            sending_receive_tracer = tracer.on_glob_subscriber_found(
                glob=subscriber.glob, url=subscriber.url
            )
            del tracer
        else:
            assert_never(subscriber)

        failed += 1
        my_identifier = secrets.token_bytes(8)
        my_tracing_and_followup = sending_receive_tracer.on_sending_receive(
            identifier=my_identifier
        )
        del sending_receive_tracer
        my_tracing = my_tracing_and_followup.tracing
        on_received_response_tracer = my_tracing_and_followup.followup
        my_authorization = await config.authorize_receive(
            tracing=my_tracing,
            url=subscriber.url,
            topic=topic,
            message_sha512=sha512,
            identifier=my_identifier,
            now=time.time(),
        )
        if my_authorization is None:
            headers.pop("Authorization", None)
        else:
            headers["Authorization"] = my_authorization

        data.seek(message_starts_at)
        try:
            raw_resp = session.post(
                subscriber.url,
                # requires monkey patching for reuse
                # https://github.com/aio-libs/aiohttp/issues/10325
                data=guarded_request_body,
                headers=headers,
            )
            resp = await raw_resp.__aenter__()
        except aiohttp.ClientError:
            handle_missed_start_tracer = on_received_response_tracer.on_network_error()
            del on_received_response_tracer
            logging.error(
                f"Failed to notify {subscriber.url} about {topic!r}", exc_info=True
            )
            tracer = await _handle_missed(
                config, topic, subscriber, handle_missed_start_tracer
            )
            continue

        on_auth_result_tracer = on_received_response_tracer.on_response_received(
            status_code=resp.status
        )
        del on_received_response_tracer

        try:
            try:
                parsed_resp = await _parse_receive_response(resp.content)
            finally:
                await raw_resp.__aexit__(None, None, None)
        except Exception:
            logging.error(
                f"Failed to notify {subscriber.url} about {topic!r}", exc_info=True
            )
            handle_missed_start_tracer = on_auth_result_tracer.on_bad_receive_response()
            del on_auth_result_tracer
            tracer = await _handle_missed(
                config, topic, subscriber, handle_missed_start_tracer
            )
            del handle_missed_start_tracer
            continue

        if (
            parsed_resp.type
            == SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
        ):
            on_unsubscribe_immediate_done_tracer = (
                on_auth_result_tracer.on_unsubscribe_immediate_requested()
            )
            del on_auth_result_tracer

            if subscriber.type == SubscriberInfoType.EXACT:
                db_result = await config.unsubscribe_exact(
                    url=subscriber.url, exact=topic
                )
            elif subscriber.type == SubscriberInfoType.GLOB:
                db_result = await config.unsubscribe_glob(
                    url=subscriber.url, glob=subscriber.glob
                )
            else:
                assert_never(subscriber)

            if db_result == "unavailable":
                on_unsubscribe_immediate_done_tracer.on_unsubscribe_immediate_unavailable()
                raise Exception("db unavailable to process unsubscribe immediate")

            if db_result == "not_found":
                tracer = (
                    on_unsubscribe_immediate_done_tracer.on_unsubscribe_immediate_not_found()
                )
                continue

            if db_result != "success":
                assert_never(db_result)

            tracer = (
                on_unsubscribe_immediate_done_tracer.on_unsubscribe_immediate_success()
            )
            continue

        if (
            parsed_resp.type
            != SubscriberToBroadcasterStatelessMessageType.RESPONSE_CONFIRM_RECEIVE
        ):
            assert_never(parsed_resp)

        auth_result = await config.is_confirm_receive_allowed(
            tracing=parsed_resp.tracing,
            identifier=parsed_resp.identifier,
            num_subscribers=parsed_resp.num_subscribers,
            url=subscriber.url,
            now=time.time(),
            authorization=parsed_resp.authorization,
        )

        if auth_result != AuthResult.OK:
            handle_missed_start_tracer = (
                on_auth_result_tracer.on_bad_receive_auth_result(result=auth_result)
            )
            del on_auth_result_tracer
            tracer = await _handle_missed(
                config, topic, subscriber, handle_missed_start_tracer
            )
            del handle_missed_start_tracer
            continue

        tracer = on_auth_result_tracer.on_receive_confirmed(
            tracing=parsed_resp.tracing,
            num_subscribers=parsed_resp.num_subscribers,
        )
        del on_auth_result_tracer

        failed -= 1
        succeeded += 1

    final_tracer = tracer.on_no_more_subscribers()
    del tracer

    return TrustedNotifyResultOK(
        type=TrustedNotifyResultType.OK,
        succeeded=succeeded,
        failed=failed,
        tracer=final_tracer,
    )


@fast_dataclass
class _ReceiveResponseUnsubscribeImmediate:
    type: Literal[
        SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
    ]


@fast_dataclass
class _ReceiveResponseConfirmReceive:
    type: Literal[SubscriberToBroadcasterStatelessMessageType.RESPONSE_CONFIRM_RECEIVE]
    authorization: Optional[str]
    tracing: bytes
    identifier: bytes
    num_subscribers: int


async def _parse_receive_response(
    rdr: AsyncReadableBytesIO,
) -> Union[_ReceiveResponseUnsubscribeImmediate, _ReceiveResponseConfirmReceive]:
    """Raises ValueError typically if the body is bad"""
    header_type_bytes = await async_read_exact(rdr, 2)
    header_type = int.from_bytes(header_type_bytes, "big")

    if header_type == int(
        SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
    ):
        return _ReceiveResponseUnsubscribeImmediate(
            type=SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
        )

    if header_type != int(
        SubscriberToBroadcasterStatelessMessageType.RESPONSE_CONFIRM_RECEIVE
    ):
        raise ValueError(f"unexpected response first 2 bytes: {header_type!r}")

    authorization_length_bytes = await async_read_exact(rdr, 2)
    authorization_length = int.from_bytes(authorization_length_bytes, "big")
    authorization_bytes = await async_read_exact(rdr, authorization_length)
    authorization = (
        authorization_bytes.decode("utf-8") if authorization_bytes != b"" else None
    )

    tracing_length_bytes = await async_read_exact(rdr, 2)
    tracing_length = int.from_bytes(tracing_length_bytes, "big")
    tracing = await async_read_exact(rdr, tracing_length)

    identifier_length_bytes = await async_read_exact(rdr, 1)
    identifier_length = int.from_bytes(identifier_length_bytes, "big")
    identifier = await async_read_exact(rdr, identifier_length)

    num_subscribers_bytes = await async_read_exact(rdr, 4)
    num_subscribers = int.from_bytes(num_subscribers_bytes, "big")
    return _ReceiveResponseConfirmReceive(
        type=SubscriberToBroadcasterStatelessMessageType.RESPONSE_CONFIRM_RECEIVE,
        authorization=authorization,
        tracing=tracing,
        identifier=identifier,
        num_subscribers=num_subscribers,
    )



import asyncio
import dataclasses
import hashlib
import io
import tempfile
import time
from typing import IO, cast

from lonelypsp.auth.config import AuthResult
from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType
from lonelypsp.stateful.messages.confirm_notify import (
    B2S_ConfirmNotify,
    serialize_b2s_confirm_notify,
)
from lonelypsp.stateful.messages.continue_notify import (
    B2S_ContinueNotify,
    serialize_b2s_continue_notify,
)
from lonelypsp.stateful.messages.notify_stream import S2B_NotifyStream
from lonelypsp.tracing.impl.noop.shared.handle_trusted_notify import (
    NoopHandleTrustedNotify,
)

from lonelypss.routes.notify import TrustedNotifyResultType, handle_trusted_notify
from lonelypss.ws.handlers.open.collector_utils import (
    maybe_write_large_message_for_training,
)
from lonelypss.ws.handlers.open.compressor_utils import reserve_decompressor
from lonelypss.ws.handlers.open.errors import AuthRejectedException
from lonelypss.ws.handlers.open.send_simple_asap import send_simple_asap
from lonelypss.ws.state import CompressorState, NotifyStreamState, StateOpen


async def process_notify_stream(state: StateOpen, message: S2B_NotifyStream) -> None:
    """Processes a request by the subscriber to notify subscribers to a given
    topic with the given data, where that data may be sent over multiple websocket
    messages
    """

    if message.part_id is None:
        if state.notify_stream_state is not None:
            raise Exception(
                "notify stream: already in progress despite first part received"
            )

        state.notify_stream_state = NotifyStreamState(
            identifier=message.identifier,
            first=dataclasses.replace(message, payload=b""),
            part_id=-1,
            body_hasher=hashlib.sha512(),
            body=tempfile.SpooledTemporaryFile(
                max_size=state.broadcaster_config.message_body_spool_size
            ),
        )

    if state.notify_stream_state is None:
        raise Exception("notify stream: no first part received")

    if message.identifier != state.notify_stream_state.identifier:
        raise Exception("notify stream: identifier mismatch")

    received_part_id = 0 if message.part_id is None else message.part_id
    if received_part_id != state.notify_stream_state.part_id + 1:
        raise Exception("notify stream: part_id mismatch")

    first = state.notify_stream_state.first

    auth_at = time.time()
    auth_result = await state.broadcaster_config.is_notify_allowed(
        tracing=first.tracing,
        topic=first.topic,
        identifier=first.identifier,
        message_sha512=(
            first.unverified_compressed_sha512
            if first.compressor_id is not None
            else first.unverified_uncompressed_sha512
        ),
        now=auth_at,
        authorization=message.authorization,
    )

    if auth_result != AuthResult.OK:
        raise AuthRejectedException(f"notify stream: {auth_result}")

    state.notify_stream_state.body_hasher.update(message.payload)
    state.notify_stream_state.body.write(message.payload)
    state.notify_stream_state.part_id = received_part_id

    read_so_far = state.notify_stream_state.body.tell()
    expected_length = (
        first.compressed_length
        if first.compressor_id is not None
        else first.uncompressed_length
    )

    if read_so_far > expected_length:
        raise Exception("notify stream: received too much data")

    if read_so_far < expected_length:
        # TODO: cannot use a url here as we would need to ensure no other url is created
        # until the send is actually queued, which requires being in a send_task, not the
        # process_task

        resp_tracing = b""  # TODO: tracing
        resp_authorization = (
            await state.broadcaster_config.authorize_stateful_continue_notify(
                tracing=resp_tracing,
                identifier=first.identifier,
                part_id=received_part_id,
                now=time.time(),
            )
        )
        send_simple_asap(
            state,
            serialize_b2s_continue_notify(
                B2S_ContinueNotify(
                    type=BroadcasterToSubscriberStatefulMessageType.CONTINUE_NOTIFY,
                    identifier=message.identifier,
                    part_id=received_part_id,
                    authorization=resp_authorization,
                    tracing=resp_tracing,
                ),
                minimal_headers=state.broadcaster_config.websocket_minimal_headers,
            ),
        )
        return

    actual_sha512 = state.notify_stream_state.body_hasher.digest()
    expected_sha512 = (
        first.unverified_compressed_sha512
        if first.compressor_id is not None
        else first.unverified_uncompressed_sha512
    )

    if actual_sha512 != expected_sha512:
        raise Exception("notify stream: integrity check failed (hash mismatch)")

    if expected_length == 0:
        # different strategy in this case to avoid read(0)
        if first.compressor_id is not None:
            raise Exception("notify stream: empty compressed message")

        notify_result = await handle_trusted_notify(
            first.topic,
            io.BytesIO(b""),
            config=state.broadcaster_config,
            session=state.client_session,
            content_length=0,
            sha512=actual_sha512,
            tracer=NoopHandleTrustedNotify(),  # TODO: tracing
        )
        if notify_result.type != TrustedNotifyResultType.OK:
            raise Exception(f"notify stream failed: {notify_result}")

        resp_tracing = b""  # TODO: tracing
        resp_authorization = await state.broadcaster_config.authorize_confirm_notify(
            tracing=resp_tracing,
            identifier=first.identifier,
            subscribers=notify_result.succeeded,
            topic=first.topic,
            message_sha512=actual_sha512,
            now=time.time(),
        )
        send_simple_asap(
            state,
            serialize_b2s_confirm_notify(
                B2S_ConfirmNotify(
                    type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY,
                    identifier=message.identifier,
                    subscribers=notify_result.succeeded,
                    authorization=resp_authorization,
                    tracing=resp_tracing,
                ),
                minimal_headers=state.broadcaster_config.websocket_minimal_headers,
            ),
        )
        state.notify_stream_state.body.close()
        state.notify_stream_state = None
        return

    body = state.notify_stream_state.body
    if first.compressor_id is None:
        with maybe_write_large_message_for_training(
            state, first.uncompressed_length
        ) as training_writer:
            if training_writer.is_void:
                training_writer.skip_void()
            else:
                body.seek(0)

                while True:
                    chunk = body.read(
                        min(io.DEFAULT_BUFFER_SIZE, training_writer.remaining)
                    )
                    if not chunk:
                        break
                    training_writer.write_chunk(chunk)
                    if not training_writer.remaining:
                        break

                    await asyncio.sleep(0)

        body.seek(0)
        notify_result = await handle_trusted_notify(
            first.topic,
            body,
            config=state.broadcaster_config,
            session=state.client_session,
            content_length=first.uncompressed_length,
            sha512=actual_sha512,
            tracer=NoopHandleTrustedNotify(),  # TODO: tracing
        )

        if notify_result.type != TrustedNotifyResultType.OK:
            raise Exception(f"notify stream failed: {notify_result}")

        resp_tracing = b""  # TODO: tracing
        resp_authorization = await state.broadcaster_config.authorize_confirm_notify(
            tracing=resp_tracing,
            identifier=first.identifier,
            subscribers=notify_result.succeeded,
            topic=first.topic,
            message_sha512=actual_sha512,
            now=time.time(),
        )
        send_simple_asap(
            state,
            serialize_b2s_confirm_notify(
                B2S_ConfirmNotify(
                    type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY,
                    identifier=message.identifier,
                    subscribers=notify_result.succeeded,
                    authorization=resp_authorization,
                    tracing=resp_tracing,
                ),
                minimal_headers=state.broadcaster_config.websocket_minimal_headers,
            ),
        )
        body.close()
        state.notify_stream_state = None
        return

    for candidate_compressor in state.compressors:
        if candidate_compressor.identifier == first.compressor_id:
            compressor = candidate_compressor
            break
    else:
        raise ValueError(f"compressor not found: {first.compressor_id}")

    if compressor.type == CompressorState.PREPARING:
        compressor = await compressor.task

    with tempfile.SpooledTemporaryFile(
        max_size=state.broadcaster_config.message_body_spool_size
    ) as decompressed_file:
        hasher = hashlib.sha512()
        pos = 0
        body.seek(0)
        with (
            maybe_write_large_message_for_training(
                state, first.decompressed_length
            ) as training_writer,
            reserve_decompressor(state, compressor) as reserved_decompressor,
            reserved_decompressor.stream_reader(cast(IO[bytes], body)) as streamer,
        ):
            while True:
                chunk = streamer.read(io.DEFAULT_BUFFER_SIZE)
                if not chunk:
                    break

                pos += len(chunk)
                if pos > first.decompressed_length:
                    raise Exception("notify stream: received too much data")

                decompressed_file.write(chunk)
                training_writer.write_chunk(chunk)
                hasher.update(chunk)

                await asyncio.sleep(0)

        if pos != first.decompressed_length:
            raise Exception("notify stream: received too little data")

        decompressed_sha512 = hasher.digest()

        state.notify_stream_state.body.close()
        state.notify_stream_state = None

        decompressed_file.seek(0)
        notify_result = await handle_trusted_notify(
            first.topic,
            decompressed_file,
            config=state.broadcaster_config,
            session=state.client_session,
            content_length=first.decompressed_length,
            sha512=decompressed_sha512,
            tracer=NoopHandleTrustedNotify(),  # TODO: tracing
        )
        if notify_result.type != TrustedNotifyResultType.OK:
            raise Exception(f"notify stream failed: {notify_result}")

        resp_tracing = b""  # TODO: tracing
        resp_authorization = await state.broadcaster_config.authorize_confirm_notify(
            tracing=resp_tracing,
            identifier=first.identifier,
            subscribers=notify_result.succeeded,
            topic=first.topic,
            message_sha512=actual_sha512,
            now=time.time(),
        )
        send_simple_asap(
            state,
            serialize_b2s_confirm_notify(
                B2S_ConfirmNotify(
                    type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY,
                    identifier=message.identifier,
                    subscribers=notify_result.succeeded,
                    authorization=resp_authorization,
                    tracing=resp_tracing,
                ),
                minimal_headers=state.broadcaster_config.websocket_minimal_headers,
            ),
        )



import asyncio
import hashlib
import io
import tempfile
import time

from lonelypsp.auth.config import AuthResult
from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType
from lonelypsp.stateful.messages.confirm_notify import (
    B2S_ConfirmNotify,
    serialize_b2s_confirm_notify,
)
from lonelypsp.stateful.messages.notify import S2B_Notify
from lonelypsp.tracing.impl.noop.shared.handle_trusted_notify import (
    NoopHandleTrustedNotify,
)

from lonelypss.routes.notify import TrustedNotifyResultType, handle_trusted_notify
from lonelypss.ws.handlers.open.collector_utils import (
    maybe_store_small_message_for_training,
    maybe_write_large_message_for_training,
)
from lonelypss.ws.handlers.open.compressor_utils import reserve_decompressor
from lonelypss.ws.handlers.open.errors import AuthRejectedException
from lonelypss.ws.handlers.open.send_simple_asap import send_simple_asap
from lonelypss.ws.state import CompressorState, StateOpen


async def process_notify(state: StateOpen, message: S2B_Notify) -> None:
    """Processes a request by the subscriber to notify subscribers to a given
    topic with the given data
    """
    auth_at = time.time()
    auth_result = await state.broadcaster_config.is_notify_allowed(
        tracing=message.tracing,
        topic=message.topic,
        identifier=message.identifier,
        message_sha512=(
            message.verified_compressed_sha512
            if message.compressor_id is not None
            else message.verified_uncompressed_sha512
        ),
        now=auth_at,
        authorization=message.authorization,
    )

    if auth_result != AuthResult.OK:
        raise AuthRejectedException(f"notify: {auth_result}")

    if message.compressor_id is None:
        maybe_store_small_message_for_training(state, message.uncompressed_message)
        notify_result = await handle_trusted_notify(
            message.topic,
            io.BytesIO(message.uncompressed_message),
            config=state.broadcaster_config,
            session=state.client_session,
            content_length=len(message.uncompressed_message),
            sha512=message.verified_uncompressed_sha512,
            tracer=NoopHandleTrustedNotify(),  # TODO: tracing
        )

        if notify_result.type != TrustedNotifyResultType.OK:
            raise ValueError(f"notify failed: {notify_result}")

        resp_tracing = b""  # TODO: tracing
        resp_authorization = await state.broadcaster_config.authorize_confirm_notify(
            tracing=resp_tracing,
            identifier=message.identifier,
            subscribers=notify_result.succeeded,
            topic=message.topic,
            message_sha512=message.verified_uncompressed_sha512,
            now=time.time(),
        )
        send_simple_asap(
            state,
            serialize_b2s_confirm_notify(
                B2S_ConfirmNotify(
                    type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY,
                    identifier=message.identifier,
                    subscribers=notify_result.succeeded,
                    authorization=resp_authorization,
                    tracing=resp_tracing,
                ),
                minimal_headers=state.broadcaster_config.websocket_minimal_headers,
            ),
        )
        return

    for candidate_compressor in state.compressors:
        if candidate_compressor.identifier == message.compressor_id:
            compressor = candidate_compressor
            break
    else:
        raise ValueError(f"compressor not found: {message.compressor_id}")

    if compressor.type == CompressorState.PREPARING:
        compressor = await compressor.task

    # although the compressed data fits in memory, we may not want to have the
    # entire uncompressed data in memory. it would neat and efficient to simply
    # decompress in parts as we are notifying subscribers, but is a bit
    # challenging to implement

    with (
        tempfile.SpooledTemporaryFile(
            max_size=state.broadcaster_config.message_body_spool_size
        ) as decompressed_file,
    ):
        pos = 0
        hasher = hashlib.sha512()

        with (
            maybe_write_large_message_for_training(
                state, message.decompressed_length
            ) as training_writer,
            reserve_decompressor(state, compressor) as reserved_decompressor,
            reserved_decompressor.stream_reader(message.compressed_message) as streamer,
        ):
            while True:
                chunk = streamer.read(io.DEFAULT_BUFFER_SIZE)
                if not chunk:
                    break
                pos += len(chunk)
                if pos > message.decompressed_length:
                    raise ValueError(
                        "decompressed length exceeded during decompression"
                    )

                decompressed_file.write(chunk)
                training_writer.write_chunk(chunk)
                hasher.update(chunk)

                await asyncio.sleep(0)

        if pos != message.decompressed_length:
            raise ValueError("decompressed length not reached during decompression")

        decompressed_sha512 = hasher.digest()
        decompressed_file.seek(0)

        notify_result = await handle_trusted_notify(
            message.topic,
            decompressed_file,
            config=state.broadcaster_config,
            session=state.client_session,
            content_length=message.decompressed_length,
            sha512=decompressed_sha512,
            tracer=NoopHandleTrustedNotify(),  # TODO: tracing
        )
        if notify_result.type != TrustedNotifyResultType.OK:
            raise ValueError(f"notify failed: {notify_result}")

        resp_tracing = b""  # TODO: tracing
        resp_authorization = await state.broadcaster_config.authorize_confirm_notify(
            tracing=resp_tracing,
            identifier=message.identifier,
            subscribers=notify_result.succeeded,
            topic=message.topic,
            message_sha512=decompressed_sha512,
            now=time.time(),
        )
        send_simple_asap(
            state,
            serialize_b2s_confirm_notify(
                B2S_ConfirmNotify(
                    type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY,
                    identifier=message.identifier,
                    subscribers=notify_result.succeeded,
                    authorization=resp_authorization,
                    tracing=resp_tracing,
                ),
                minimal_headers=state.broadcaster_config.websocket_minimal_headers,
            ),
        )



from typing import Generic, Protocol, TypeVar

from lonelypss.ws.state import StateOpen

T_contra = TypeVar("T_contra", contravariant=True)


class S2B_MessageProcessor(Generic[T_contra], Protocol):
    """The protocol that the exported functions in this folder adhere to,
    checked to ensure consistency
    """

    async def __call__(self, state: StateOpen, message: T_contra) -> None:
        """Processes the given message received while in the given state. There
        will be no other messages being processed at the same time, but this does
        not give you any exclusive access to the state as send tasks have overlapping
        responsibilities (e.g., writing to the training data collector). Thus, this
        must be async-safe.
        """



import io
from typing import TYPE_CHECKING

from lonelypss.ws.handlers.open.collector_utils import (
    maybe_store_small_message_for_training,
)
from lonelypss.ws.handlers.open.send_receive_stream import (
    send_receive_stream,
)
from lonelypss.ws.handlers.open.senders.protocol import Sender
from lonelypss.ws.state import (
    InternalSmallMessage,
    StateOpen,
)


async def send_internal_small_message(
    state: StateOpen, message: InternalSmallMessage
) -> None:
    """Notifies the client about the small message that was sent to a topic
    they are subscribed to. Note that although the broadcaster kept the
    entire message in memory it may still be necessary to split the message
    when sending over the websocket
    """
    maybe_store_small_message_for_training(state, message.data)
    await send_receive_stream(
        state,
        io.BytesIO(message.data),
        topic=message.topic,
        uncompressed_sha512=message.sha512,
        uncompressed_length=len(message.data),
        maybe_store_for_training=False,
        read_lock=None,
    )


if TYPE_CHECKING:
    _: Sender[InternalSmallMessage] = send_internal_small_message



import asyncio
import hashlib
import io
import secrets
import tempfile
import time
from typing import Optional

from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.stateful.messages.confirm_receive import S2B_ConfirmReceive
from lonelypsp.stateful.messages.continue_receive import S2B_ContinueReceive
from lonelypsp.stateful.messages.receive_stream import (
    B2S_ReceiveStreamContinuation,
    B2S_ReceiveStreamStartCompressed,
    B2S_ReceiveStreamStartUncompressed,
    serialize_b2s_receive_stream,
)

from lonelypss.util.sync_io import SyncReadableBytesIO, read_exact
from lonelypss.ws.handlers.open.collector_utils import (
    maybe_write_large_message_for_training,
)
from lonelypss.ws.handlers.open.compressor_utils import (
    choose_compressor_for_compression,
    reserve_compressor,
)
from lonelypss.ws.handlers.open.websocket_url import (
    make_for_send_websocket_url_and_change_counter,
)
from lonelypss.ws.state import (
    CompressorReady,
    StateOpen,
)


async def send_receive_stream(
    state: StateOpen,
    uncompressed_stream: SyncReadableBytesIO,
    /,
    *,
    topic: bytes,
    uncompressed_sha512: bytes,
    uncompressed_length: int,
    maybe_store_for_training: bool,
    read_lock: Optional[asyncio.Lock],
) -> None:
    """
    Sends a stream of data via potentially multiple RECEIVE_STREAM messages.
    This will determine if compression is appropriate, and if so, will compress
    the data before sending it.

    Args:
        state (StateOpen): the state
        uncompressed_stream (SyncReadableBytesIO): the readable stream of data to send
        topic (bytes): the topic for the message
        uncompressed_sha512 (bytes): the sha512 of the message
        uncompressed_length (int): the length of the stream; we assume the stream will not over-read
        maybe_store_for_training (bool): if true, if a message of this length would be
            useful for training a compression dictionary, it will be stored for that purpose,
            otherwise it will not be stored for that purpose
        read_lock (Optional[asyncio.Lock]): if not None, this lock will be acquired before
            reading from the stream and released right after reading the stream, which
            can be used to "promote" the read to async
    """
    compressor = choose_compressor_for_compression(state, uncompressed_length)

    if compressor is None:
        await send_uncompressed_receive_stream(
            state,
            uncompressed_stream,
            topic=topic,
            sha512=uncompressed_sha512,
            length=uncompressed_length,
            maybe_store_for_training=maybe_store_for_training,
            read_lock=read_lock,
        )
        return

    await send_compressed_receive_stream(
        state,
        uncompressed_stream,
        topic=topic,
        uncompressed_length=uncompressed_length,
        compressor=compressor,
        maybe_store_for_training=maybe_store_for_training,
        read_lock=read_lock,
    )


async def send_uncompressed_receive_stream(
    state: StateOpen,
    stream: SyncReadableBytesIO,
    /,
    *,
    topic: bytes,
    sha512: bytes,
    length: int,
    maybe_store_for_training: bool,
    read_lock: Optional[asyncio.Lock],
) -> None:
    """Sends an uncompressed stream of data via potentially multiple RECEIVE_STREAM
    messages

    Args:
        state (StateOpen): the state, for grabbing configuration options
        stream (SyncIOBaseLikeIO): the readable stream of data to send
        topic (bytes): the topic for the message
        sha512 (bytes): the sha512 of the message
        length (int): the length of the stream; we assume the stream will not over-read
        maybe_store_for_training (bool): if true, if a message of this length would be
            useful for training a compression dictionary, it will be stored for that purpose,
            otherwise it will not be stored for that purpose
        read_lock (Optional[asyncio.Lock]): if not None, this lock will be acquired before
            reading from the stream and released right after reading the stream, which
            can be used to "promote" the read to async
    """
    identifier = secrets.token_bytes(4)
    tracing = b""  # TODO: tracing
    authorization = await state.broadcaster_config.authorize_receive(
        tracing=tracing,
        url=make_for_send_websocket_url_and_change_counter(state),
        topic=topic,
        message_sha512=sha512,
        identifier=identifier,
        now=time.time(),
    )

    headers = serialize_b2s_receive_stream(
        B2S_ReceiveStreamStartUncompressed(
            type=BroadcasterToSubscriberStatefulMessageType.RECEIVE_STREAM,
            authorization=authorization,
            tracing=tracing,
            identifier=identifier,
            part_id=None,
            topic=topic,
            compressor_id=None,
            uncompressed_length=length,
            unverified_uncompressed_sha512=sha512,
            payload=b"",
        ),
        minimal_headers=state.broadcaster_config.websocket_minimal_headers,
    )
    await send_receive_stream_given_first_headers(
        state,
        stream,
        length,
        identifier,
        topic,
        sha512,
        headers,
        maybe_store_for_training=maybe_store_for_training,
        read_lock=read_lock,
    )


async def send_compressed_receive_stream(
    state: StateOpen,
    uncompressed_stream: SyncReadableBytesIO,
    /,
    *,
    topic: bytes,
    uncompressed_length: int,
    compressor: CompressorReady,
    maybe_store_for_training: bool,
    read_lock: Optional[asyncio.Lock],
) -> None:
    """Sends a compressed stream of data via potentially multiple RECEIVE_STREAM
    messages

    Args:
        state (StateOpen): the state, for grabbing configuration options
        uncompressed_stream (SyncIOBaseLikeIO): the readable stream of data to send
        topic (bytes): the topic for the message
        uncompressed_length (int): the length of the uncompressed stream; we assume the stream
            will not over-read
        compressor (CompressorReady): the compressor to use for compressing the data
        maybe_store_for_training (bool): if true, if a message of this length would be
            useful for training a compression dictionary, it will be stored for that purpose,
            otherwise it will not be stored for that purpose.
        read_lock (Optional[asyncio.Lock]): if not None, this lock will be acquired before
            reading from the stream and released right after reading the stream, which
            can be used to "promote" the read to async

    NOTE:
        This correctly stores the uncompressed data for training purposes, not the
        compressed data, when appropriate
    """

    with tempfile.SpooledTemporaryFile(
        max_size=state.broadcaster_config.message_body_spool_size
    ) as to_send:
        hasher = hashlib.sha512()
        with (
            maybe_write_large_message_for_training(
                state, uncompressed_length, never_store=not maybe_store_for_training
            ) as training_writer,
            reserve_compressor(state, compressor) as reserved_compressor,
        ):
            chunker = reserved_compressor.chunker(
                size=uncompressed_length, chunk_size=io.DEFAULT_BUFFER_SIZE
            )

            while True:
                if read_lock is None:
                    uncompressed_chunk = uncompressed_stream.read(
                        io.DEFAULT_BUFFER_SIZE
                    )
                else:
                    async with read_lock:
                        uncompressed_chunk = uncompressed_stream.read(
                            io.DEFAULT_BUFFER_SIZE
                        )
                if not uncompressed_chunk:
                    break

                training_writer.write_chunk(uncompressed_chunk)
                for chunk in chunker.compress(uncompressed_chunk):
                    to_send.write(chunk)
                    hasher.update(chunk)
                    await asyncio.sleep(0)

            for chunk in chunker.finish():
                to_send.write(chunk)
                hasher.update(chunk)
                await asyncio.sleep(0)

        compressed_length = to_send.tell()
        compressed_sha512 = hasher.digest()
        to_send.seek(0)

        identifier = secrets.token_bytes(4)
        tracing = b""  # TODO: tracing
        headers = serialize_b2s_receive_stream(
            B2S_ReceiveStreamStartCompressed(
                type=BroadcasterToSubscriberStatefulMessageType.RECEIVE_STREAM,
                authorization=await state.broadcaster_config.authorize_receive(
                    tracing=tracing,
                    url=make_for_send_websocket_url_and_change_counter(state),
                    topic=topic,
                    message_sha512=compressed_sha512,
                    identifier=identifier,
                    now=time.time(),
                ),
                tracing=tracing,
                identifier=identifier,
                part_id=None,
                topic=topic,
                compressor_id=compressor.identifier,
                compressed_length=compressed_length,
                decompressed_length=uncompressed_length,
                unverified_compressed_sha512=compressed_sha512,
                payload=b"",
            ),
            minimal_headers=state.broadcaster_config.websocket_minimal_headers,
        )
        await send_receive_stream_given_first_headers(
            state,
            to_send,
            compressed_length,
            identifier,
            topic,
            compressed_sha512,
            headers,
            maybe_store_for_training=False,
            read_lock=None,
        )


async def send_receive_stream_given_first_headers(
    state: StateOpen,
    stream: SyncReadableBytesIO,
    length: int,
    identifier: bytes,
    topic: bytes,
    sha512: bytes,
    first_headers: bytes,
    maybe_store_for_training: bool,
    read_lock: Optional[asyncio.Lock],
) -> None:
    """Sends the given stream of data to the subscriber via potentially multiple
    RECEIVE_STREAM messages, using the given headers for the first message.

    Args:
        state (StateOpen): the state, for grabbing configuration options
        stream (SyncReadableBytesIO): the readable stream of data to send
        length (int): the length of the stream; we assume the stream will not over-read
        identifier (bytes): the identifier for the message
        topic (bytes): the topic for the message
        sha512 (bytes): the sha512 of the message
        first_headers (bytes): the headers for the first message
        maybe_store_for_training (bool): if true, if a message of this length would be
            useful for training a compression dictionary, it will be stored for that purpose,
            otherwise it will not be stored for that purpose.
        read_lock (Optional[asyncio.Lock]): if not None, this lock will be acquired before
            reading from the stream and released right after reading the stream, which
            can be used to "promote" the read to async

    NOTE:
        This only handles the data from the stream as-is, meaning that if
        `maybe_store_for_training` is true, then the stream should be for
        uncompressed data.
    """
    headers = first_headers
    msg_size = state.broadcaster_config.outgoing_max_ws_message_size or (2**64 - 1)

    part_id = 0
    pos = 0
    with maybe_write_large_message_for_training(
        state, length, never_store=not maybe_store_for_training
    ) as training_writer:
        while True:
            target_read_amount = min(length - pos, max(512, msg_size - len(headers)))

            if target_read_amount == 0:
                payload = b""
            elif read_lock is None:
                payload = read_exact(stream, target_read_amount)
            else:
                async with read_lock:
                    payload = read_exact(stream, target_read_amount)
            training_writer.write_chunk(payload)
            pos += len(payload)
            is_done = pos >= length
            await state.expecting_acks.put(
                S2B_ConfirmReceive(
                    type=SubscriberToBroadcasterStatefulMessageType.CONFIRM_RECEIVE,
                    identifier=identifier,
                    authorization=None,
                    tracing=b"",
                    num_subscribers=1,
                )
                if is_done
                else S2B_ContinueReceive(
                    type=SubscriberToBroadcasterStatefulMessageType.CONTINUE_RECEIVE,
                    identifier=identifier,
                    part_id=part_id,
                    authorization=None,
                    tracing=b"",
                )
            )
            await state.websocket.send_bytes(headers + payload)

            if is_done:
                return

            part_id += 1
            tracing = b""  # TODO: tracing
            authorization = await state.broadcaster_config.authorize_receive(
                tracing=tracing,
                url=make_for_send_websocket_url_and_change_counter(state),
                topic=topic,
                message_sha512=sha512,
                identifier=identifier,
                now=time.time(),
            )
            headers = serialize_b2s_receive_stream(
                B2S_ReceiveStreamContinuation(
                    type=BroadcasterToSubscriberStatefulMessageType.RECEIVE_STREAM,
                    authorization=authorization,
                    tracing=tracing,
                    identifier=identifier,
                    part_id=part_id,
                    payload=b"",
                ),
                minimal_headers=state.broadcaster_config.websocket_minimal_headers,
            )



from typing import TYPE_CHECKING

from lonelypss.ws.handlers.open.send_receive_stream import send_receive_stream
from lonelypss.ws.handlers.open.senders.protocol import Sender
from lonelypss.ws.state import StateOpen, WaitingInternalSpooledLargeMessage


async def send_waiting_internal_spooled_large_message(
    state: StateOpen, message: WaitingInternalSpooledLargeMessage
) -> None:
    """Sends the message to the ASGI server for forwarding"""
    try:
        await send_receive_stream(
            state,
            message.stream,
            topic=message.topic,
            uncompressed_sha512=message.sha512,
            uncompressed_length=message.length,
            maybe_store_for_training=True,
            read_lock=None,
        )
    finally:
        message.stream.close()


if TYPE_CHECKING:
    _: Sender[WaitingInternalSpooledLargeMessage] = (
        send_waiting_internal_spooled_large_message
    )



import sys
from typing import List, Union, cast


class NormalDisconnectException(Exception): ...


class AuthRejectedException(Exception): ...


if sys.version_info < (3, 11):

    def combine_multiple_base_exceptions(
        msg: str, excs: List[BaseException]
    ) -> BaseException:
        """Raises a single BaseException whose __cause__ includes all
        the indicate exceptions and their causes.
        """
        if not excs:
            raise ValueError("no exceptions to combine")

        if len(excs) == 1:
            return excs[0]

        exc = BaseException(msg)
        last_exc = exc

        for nexc in excs:
            while last_exc.__cause__ is not None:
                last_exc = last_exc.__cause__
            last_exc.__cause__ = nexc

        return exc

    def combine_multiple_normal_exceptions(
        msg: str, excs: List[Exception]
    ) -> Exception:
        """Raises a single Exception whose __cause__ includes all
        the indicate exceptions and their causes.
        """
        if not excs:
            raise ValueError("no exceptions to combine")

        if len(excs) == 1:
            return excs[0]

        exc = Exception(msg)
        last_exc: Union[Exception, BaseException] = exc

        for nexc in excs:
            while last_exc.__cause__ is not None:
                last_exc = last_exc.__cause__
            last_exc.__cause__ = nexc

        return exc

else:

    def combine_multiple_base_exceptions(
        msg: str, excs: List[BaseException]
    ) -> BaseException:
        """Light wrapper around BaseExceptionGroup"""
        if not excs:
            raise ValueError("no exceptions to combine")

        if len(excs) == 1:
            return excs[0]

        if any(isinstance(e, BaseExceptionGroup) for e in excs):
            new_excs: List[BaseException] = []
            for e in excs:
                if isinstance(e, BaseExceptionGroup):
                    new_excs.extend(e.exceptions)
                else:
                    new_excs.append(e)
            excs = new_excs

        return BaseExceptionGroup(msg, excs)

    def combine_multiple_normal_exceptions(
        msg: str, excs: List[Exception]
    ) -> Exception:
        """Light wrapper around ExceptionGroup"""
        if not excs:
            raise ValueError("no exceptions to combine")

        if len(excs) == 1:
            return excs[0]

        if any(isinstance(e, ExceptionGroup) for e in excs):
            new_excs: List[Exception] = []
            for e in excs:
                if isinstance(e, ExceptionGroup):
                    new_excs.extend(e.exceptions)
                else:
                    new_excs.append(e)
            excs = new_excs

        return ExceptionGroup(msg, excs)


def combine_multiple_exceptions(msg: str, excs: List[BaseException]) -> BaseException:
    """Returns a single Exception, if possible, otherwise a BaseException which will
    report all the indicated exceptions in the most informative way possible.
    """
    if all(isinstance(e, Exception) for e in excs):
        return combine_multiple_normal_exceptions(msg, cast(List[Exception], excs))
    return combine_multiple_base_exceptions(msg, excs)



import asyncio
import io
import tempfile
from types import TracebackType
from typing import TYPE_CHECKING, Optional, Type

from lonelypss.util.sync_io import SyncIOBaseLikeIO, SyncReadableBytesIO
from lonelypss.ws.handlers.open.send_receive_stream import send_receive_stream
from lonelypss.ws.handlers.open.senders.protocol import Sender
from lonelypss.ws.state import InternalLargeMessage, StateOpen


async def send_internal_large_message(
    state: StateOpen, message: InternalLargeMessage
) -> None:
    """Sends the message to the ASGI server for forwarding; if it takes too
    long to send it to the ASGI server, copies it to file so that the message
    can be used by the next receiver
    """
    try:
        if message.length == 0:
            message.finished.set()
            await send_receive_stream(
                state,
                io.BytesIO(b""),
                topic=message.topic,
                uncompressed_sha512=message.sha512,
                uncompressed_length=0,
                maybe_store_for_training=False,
                read_lock=None,
            )
            return

        if state.broadcaster_config.websocket_large_direct_send_timeout is None:
            await send_receive_stream(
                state,
                message.stream,
                topic=message.topic,
                uncompressed_sha512=message.sha512,
                uncompressed_length=message.length,
                maybe_store_for_training=True,
                read_lock=None,
            )
            message.finished.set()
            return

        read_lock = asyncio.Lock()
        with _SwappableSyncReadableBytesIO(message.stream, message.length) as stream:
            timeout = asyncio.create_task(
                asyncio.sleep(
                    state.broadcaster_config.websocket_large_direct_send_timeout
                )
            )
            send_task = asyncio.create_task(
                send_receive_stream(
                    state,
                    stream,
                    topic=message.topic,
                    uncompressed_sha512=message.sha512,
                    uncompressed_length=message.length,
                    maybe_store_for_training=True,
                    read_lock=read_lock,
                )
            )

            await asyncio.wait(
                [timeout, send_task], return_when=asyncio.FIRST_COMPLETED
            )

            timeout.cancel()
            if not send_task.done():
                async with read_lock:
                    await stream.swap(
                        spool_size=state.broadcaster_config.message_body_spool_size
                    )
                message.finished.set()
                await send_task
            else:
                message.finished.set()
                send_task.result()
    except BaseException:
        message.finished.set()
        raise


class _SwappableSyncReadableBytesIO:
    def __init__(
        self, original_stream: SyncReadableBytesIO, original_length: int
    ) -> None:
        self._original_stream: Optional[SyncReadableBytesIO] = original_stream
        self._original_remaining: Optional[int] = original_length
        self._swapped_stream: Optional[SyncIOBaseLikeIO] = None

    def __enter__(self) -> "_SwappableSyncReadableBytesIO":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()

    async def swap(self, spool_size: int) -> None:
        if self._original_stream is None:
            return

        assert self._original_remaining is not None

        target = tempfile.SpooledTemporaryFile(max_size=spool_size)
        remaining = self._original_remaining
        try:
            while True:
                chunk = self._original_stream.read(
                    min(remaining, io.DEFAULT_BUFFER_SIZE)
                )
                if not chunk:
                    break
                target.write(chunk)
                await asyncio.sleep(0)

            target.seek(0)
            self._swapped_stream = target
        except BaseException:
            target.close()
            raise

    def read(self, n: int) -> bytes:
        if self._swapped_stream is not None:
            result = self._swapped_stream.read(n)
            if not result:
                self._swapped_stream.close()
                self._swapped_stream = None
            return result

        if self._original_stream is None:
            return b""

        assert self._original_remaining is not None
        chunk = self._original_stream.read(min(n, self._original_remaining))
        self._original_remaining -= len(chunk)
        if self._original_remaining == 0:
            self._original_stream = None
            self._original_remaining = None
        return chunk

    def close(self) -> None:
        self._original_stream = None
        self._original_remaining = None
        if self._swapped_stream is not None:
            self._swapped_stream.close()
            self._swapped_stream = None


if TYPE_CHECKING:
    _: Sender[InternalLargeMessage] = send_internal_large_message



import time
from typing import Annotated, Optional

from fastapi import APIRouter, Header, Request, Response
from lonelypsp.auth.config import AuthResult
from lonelypsp.compat import assert_never
from lonelypsp.stateless.constants import SubscriberToBroadcasterStatelessMessageType
from lonelypsp.sync_io import PreallocatedBytesIO

from lonelypss.middleware.config import get_config_from_request
from lonelypss.middleware.ws_receiver import get_ws_receiver_from_request
from lonelypss.util.async_io import async_read_exact
from lonelypss.util.request_body_io import AsyncIterableAIO

router = APIRouter()


@router.post(
    "/v1/missed_for_websockets",
)
async def missed_for_websockets(
    request: Request, authorization: Annotated[Optional[str], Header()] = None
) -> Response:
    """As a broadcaster, in order to handle websocket connections, we need to be
    notified about messages that were sent to other broadcasters. If the other
    broadcaster fails to notify us about a message, it is helpful for that
    broadcaster to let us know as soon as possible, without building up an
    excessive amount of state that will likely lead to cascading errors.

    This is the endpoint that broadcasters (including ourself via the shared db)
    uses to notify us that we may have missed a `receive_for_websockets` call on
    a topic we were subscribed to, so we can forward that information to the
    websocket connections to trigger the same recovery method they use if their
    websocket connection temporarily drops.
    """
    config = get_config_from_request(request)
    receiver = get_ws_receiver_from_request(request)

    if str(request.url) != receiver.missed_url:
        return Response(
            status_code=400,
            headers={"Content-Type": "application/json; charset=utf-8"},
            content=b'{"unsubscribe": true, "reason": "invalid missed URL"}',
        )

    stream = request.stream()
    try:
        body = AsyncIterableAIO(stream.__aiter__())
        tracing_length_bytes = await async_read_exact(body, 2)
        tracing_length = int.from_bytes(tracing_length_bytes, "big")
        tracing = await async_read_exact(body, tracing_length)
        topic_length_bytes = await async_read_exact(body, 2)
        topic_length = int.from_bytes(topic_length_bytes, "big")
        topic = await async_read_exact(body, topic_length)
    finally:
        await stream.aclose()

    auth_result = await config.is_missed_allowed(
        tracing=tracing,
        recovery=receiver.missed_url,
        topic=topic,
        now=time.time(),
        authorization=authorization,
    )
    if auth_result == AuthResult.UNAVAILABLE:
        return Response(status_code=503)
    if auth_result == AuthResult.FORBIDDEN:
        return Response(status_code=403)
    if auth_result == AuthResult.UNAUTHORIZED:
        return Response(status_code=401)
    if auth_result != AuthResult.OK:
        assert_never(auth_result)

    await receiver.on_missed(topic=topic)

    resp_tracing = b""  # TODO: tracing
    resp_authorization = await config.authorize_confirm_missed(
        tracing=resp_tracing,
        topic=topic,
        url=receiver.missed_url,
        now=time.time(),
    )
    resp_authorization_bytes = (
        b"" if resp_authorization is None else resp_authorization.encode("utf-8")
    )
    resp_body = PreallocatedBytesIO(
        2 + 2 + len(resp_authorization_bytes) + 2 + len(resp_tracing)
    )
    resp_body.write(
        int(
            SubscriberToBroadcasterStatelessMessageType.RESPONSE_CONFIRM_MISSED
        ).to_bytes(2, "big")
    )
    resp_body.write(len(resp_authorization_bytes).to_bytes(2, "big"))
    resp_body.write(resp_authorization_bytes)
    resp_body.write(len(resp_tracing).to_bytes(2, "big"))
    resp_body.write(resp_tracing)
    return Response(
        content=memoryview(resp_body.buffer),
        status_code=200,
        headers={"Content-Type": "application/octet-stream"},
    )



from fastapi import APIRouter, WebSocket

from lonelypss.middleware.config import get_config_from_request
from lonelypss.middleware.ws_receiver import get_ws_receiver_from_request
from lonelypss.ws.handlers.handler import handle_any
from lonelypss.ws.state import State, StateAccepting, StateType

router = APIRouter()


@router.websocket("/v1/websocket")
async def websocket_endpoint(websocket: WebSocket) -> None:
    """Allows sending and receiving notifications over a websocket connection,
    as opposed to the typical way this library is used (HTTP requests). This is
    helpful for the following scenarios:

    - You need to send a large number of notifications, OR
    - You need to receive a large number of notifications, OR
    - You need to receive notifications for a short period of time before unsubscribing, OR
    - You need to receive some notifications, but you cannot accept incoming HTTP requests

    For maximum compatibility with websocket clients, we only communicate
    over the websocket itself (not the http-level header fields).

    Each websocket message is independently authorized, which means if using a
    suitable authorization scheme (eg HMAC) and secrecy is not required, it may
    be reasonable to serve this over WS instead of WSS. It is only recommended
    that this be done on internal networks where the auth is only intended to
    mitigate vulnerability amplification.

    ## COMPRESSION

    For notifications (both posted and received) over websockets, this supports
    using zstandard compression. It will either use an embedded dictionary, a
    precomputed dictionary, or a trained dictionary. Under the typical settings, this:

    - Only considers messages that are between 32 and 16384 bytes for training
    - Will train once after 100kb of data is ready, and once more after 10mb of data is ready,
      then will sample 10mb every 24 hours
    - Will only used the trained dictionary on messages that would be used for training

    ## MESSAGES

    messages always begin as follows

    - 2 bytes (F): flags (interpret as big-endian):
        - least significant bit (1): 0 if headers are expanded, 1 if headers are minimal
    - 2 bytes (T): type of message; see below, depends on if it's sent by a subscriber
      or the broadcaster big-endian encoded, unsigned

    EXPANDED HEADERS:
        - 2 bytes (N): number of headers, big-endian encoded, unsigned
        - REPEAT N:
            - 2 bytes (M): length of header name, big-endian encoded, unsigned
            - M bytes: header name, ascii-encoded
            - 2 bytes (L): length of header value, big-endian encoded, unsigned
            - L bytes: header value

    MINIMAL HEADERS:
    the order of the headers are fixed based on the type, in the order documented.
    Given N headers:
    - Repeat N:
        - 2 bytes (L): length of header value, big-endian encoded, unsigned
        - L bytes: header value

    See the `lonelypsp` repository for precise details on the protocol, specifically,
    `lonelypsp.stateful.constants` is a good place to start
    """
    config = get_config_from_request(websocket)
    receiver = get_ws_receiver_from_request(websocket)

    state: State = StateAccepting(
        type=StateType.ACCEPTING,
        websocket=websocket,
        broadcaster_config=config,
        internal_receiver=receiver,
    )
    while state.type != StateType.CLOSED:
        state = await handle_any(state)



import time
from typing import Annotated, Optional

from fastapi import APIRouter, Header, Request
from fastapi.responses import Response
from lonelypsp.auth.config import AuthResult
from lonelypsp.compat import assert_never
from lonelypsp.stateless.constants import BroadcasterToSubscriberStatelessMessageType
from lonelypsp.sync_io import PreallocatedBytesIO

from lonelypss.middleware.config import get_config_from_request
from lonelypss.util.async_io import async_read_exact
from lonelypss.util.request_body_io import AsyncIterableAIO

router = APIRouter()


@router.post(
    "/v1/check_subscriptions",
    status_code=200,
    responses={
        401: {"description": "Authorization header is required but not provided"},
        403: {"description": "Authorization header is provided but invalid"},
        500: {"description": "Unexpected error occurred"},
        503: {"description": "Service is unavailable, try again soon"},
    },
)
async def check_subscriptions(
    request: Request, authorization: Annotated[Optional[str], Header()] = None
) -> Response:
    """Retrieves the strong etag representing the subscriptions associated with
    the URL indicated in the request body. The exact way that the etag is
    produced is in the lonelypsp documentation; it is expected that the
    subscriber uses this endpoint to confirm the subscriptions for the url
    are still as expected by computing the expected strong etag, then
    comparing it with the etag provided in the response. If incorrect, the
    subscriber is expected to use `/v1/set_subscriptions` to update the
    subscriptions.

    NOTE: this operation is not atomic; if subscriptions are updated during
    the call, then only the following is guarranteed:

    - etag includes every subscription in the db the entire call
    - etag does not include any subscription never in the db the entire call
    - etag does not include duplicates

    ### request body
    - 2 bytes (N): length of the subscriber url to check, big-endian, unsigned
    - N bytes: the url to check, utf-8 encoded

    ### response body
    - 2 bytes (type): int(RESPONSE_CHECK_SUBSCRIPTIONS), big endian, unsigned
    - 2 bytes (A): big-endian, unsigned, the length of the authorization
    - A bytes: the authorization
    - 2 bytes (T): big-endian, unsigned, the length of tracing data
    - T bytes: the tracing data
    - 1 byte (reserved for etag format): 0
    - 64 bytes: the etag
    """
    config = get_config_from_request(request)

    try:
        stream = request.stream()
        try:
            body = AsyncIterableAIO(stream.__aiter__())

            tracing_length_bytes = await async_read_exact(body, 2)
            tracing_length = int.from_bytes(tracing_length_bytes, "big")
            tracing = await async_read_exact(body, tracing_length)
            url_length_bytes = await async_read_exact(body, 2)
            url_length = int.from_bytes(url_length_bytes, "big")
            url_bytes = await async_read_exact(body, url_length)
            url = url_bytes.decode("utf-8")
        finally:
            await stream.aclose()
    except ValueError:
        return Response(status_code=400)

    auth_at = time.time()
    auth_result = await config.is_check_subscriptions_allowed(
        tracing=tracing, url=url, now=auth_at, authorization=authorization
    )

    if auth_result == AuthResult.UNAUTHORIZED:
        return Response(status_code=401)
    elif auth_result == AuthResult.FORBIDDEN:
        return Response(status_code=403)
    elif auth_result == AuthResult.UNAVAILABLE:
        return Response(status_code=503)
    elif auth_result != AuthResult.OK:
        assert_never(auth_result)

    etag = await config.check_subscriptions(url=url)

    resp_tracing = b""  # TODO: tracing
    resp_authorization = await config.authorize_check_subscriptions_response(
        tracing=resp_tracing,
        strong_etag=etag,
        now=time.time(),
    )
    resp_authorization_bytes = (
        b"" if resp_authorization is None else resp_authorization.encode("utf-8")
    )
    result = PreallocatedBytesIO(
        2
        + 2
        + len(resp_authorization_bytes)
        + 2
        + len(resp_tracing)
        + 1
        + len(etag.etag)
    )
    result.write(
        int(
            BroadcasterToSubscriberStatelessMessageType.RESPONSE_CHECK_SUBSCRIPTIONS
        ).to_bytes(2, "big")
    )
    result.write(len(resp_authorization_bytes).to_bytes(2, "big"))
    result.write(resp_authorization_bytes)
    result.write(len(resp_tracing).to_bytes(2, "big"))
    result.write(resp_tracing)
    result.write(etag.format.to_bytes(1, "big"))
    result.write(etag.etag)
    return Response(
        content=memoryview(result.buffer),
        headers={"Content-Type": "application/octet-stream"},
        status_code=200,
    )



from fastapi.requests import HTTPConnection
from starlette.types import ASGIApp, Receive, Scope, Send

from lonelypss.util.ws_receiver import FanoutWSReceiver


class WSReceiverMiddleware:
    """Injects `lonelypss_ws_receiver` into the scope of the request. Required for
    websocket related routes (receive_for_websockets, websocket_endpoint) to function
    correctly
    """

    def __init__(self, app: ASGIApp, ws_receiver: FanoutWSReceiver):
        self.app = app
        self.ws_receiver = ws_receiver

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["lonelypss_ws_receiver"] = self.ws_receiver
        await self.app(scope, receive, send)


def get_ws_receiver_from_request(request: HTTPConnection) -> FanoutWSReceiver:
    """Retrieves the `lonelypss_ws_receiver` from the request scope. This is only
    available if the `WSReceiverMiddleware` is being used
    """
    try:
        return request.scope["lonelypss_ws_receiver"]
    except KeyError:
        raise Exception("WSReceiverMiddleware not injecting into request scope")



from typing import TYPE_CHECKING, cast

from lonelypsp.stateful.constants import SubscriberToBroadcasterStatefulMessageType
from lonelypsp.stateful.message import S2B_Message

from lonelypss.ws.handlers.open.processors.process_confirm_receive import (
    process_confirm_receive,
)
from lonelypss.ws.handlers.open.processors.process_continue_receive import (
    process_continue_receive,
)
from lonelypss.ws.handlers.open.processors.process_notify import process_notify
from lonelypss.ws.handlers.open.processors.process_notify_stream import (
    process_notify_stream,
)
from lonelypss.ws.handlers.open.processors.process_subscribe_exact import (
    process_subscribe_exact,
)
from lonelypss.ws.handlers.open.processors.process_subscribe_glob import (
    process_subscribe_glob,
)
from lonelypss.ws.handlers.open.processors.process_unsubscribe_exact import (
    process_unsubscribe_exact,
)
from lonelypss.ws.handlers.open.processors.process_unsubscribe_glob import (
    process_unsubscribe_glob,
)
from lonelypss.ws.handlers.open.processors.protocol import S2B_MessageProcessor
from lonelypss.ws.state import StateOpen

PROCESSORS = {
    SubscriberToBroadcasterStatefulMessageType.NOTIFY_STREAM: process_notify_stream,
    SubscriberToBroadcasterStatefulMessageType.NOTIFY: process_notify,
    SubscriberToBroadcasterStatefulMessageType.SUBSCRIBE_EXACT: process_subscribe_exact,
    SubscriberToBroadcasterStatefulMessageType.SUBSCRIBE_GLOB: process_subscribe_glob,
    SubscriberToBroadcasterStatefulMessageType.UNSUBSCRIBE_EXACT: process_unsubscribe_exact,
    SubscriberToBroadcasterStatefulMessageType.UNSUBSCRIBE_GLOB: process_unsubscribe_glob,
    SubscriberToBroadcasterStatefulMessageType.CONFIRM_RECEIVE: process_confirm_receive,
    SubscriberToBroadcasterStatefulMessageType.CONTINUE_RECEIVE: process_continue_receive,
}


async def process_any(state: StateOpen, message: S2B_Message) -> None:
    """Processes a message from the subscriber to the broadcaster. This is
    not async safe in the sense we only expect to be processing one message
    at a time per websocket in order to give a predictable ack order, but
    can handle the send task running in a separate coroutine, which may access
    some similar resources

    Raises an exception if the broadcaster should disconnect the subscriber and
    cleanup resources
    """
    await cast(S2B_MessageProcessor[S2B_Message], PROCESSORS[message.type])(
        state, message
    )


if TYPE_CHECKING:
    _: S2B_MessageProcessor[S2B_Message] = process_any



from enum import Enum, auto


class CheckResult(Enum):
    """The handler function is composed into parts; each part can either request
    we return back to the top of the handler function, continue to the next part
    of the handler function, or raise an exception to start the cleanup and disconnect
    process.

    Although the return type could be annotated as a boolean, for clarity
    we instead use an enum
    """

    RESTART = auto()
    """Return to the top of the handler function because we made progress"""

    CONTINUE = auto()
    """Continue to the next part of the handler function"""



from typing import Generic, Protocol, TypeVar

from lonelypss.ws.state import StateOpen

T_contra = TypeVar("T_contra", contravariant=True)


class Sender(Generic[T_contra], Protocol):
    """The protocol for the send_* functions within this folder; checked to
    ensure consistency
    """

    async def __call__(self, state: StateOpen, message: T_contra) -> None: ...



import asyncio

from lonelypss.ws.handlers.open.check_result import CheckResult
from lonelypss.ws.handlers.open.senders.send_any import send_any
from lonelypss.ws.state import StateOpen


async def check_send_task(state: StateOpen) -> CheckResult:
    """If the send task is unset but there are unsent messages, sets
    the send task. Otherwise, if the send task has completed, handles the result.

    Raises an exception to indicate we should move to the cleanup and disconnect
    process
    """
    if state.send_task is None:
        if state.unsent_messages:
            state.send_task = asyncio.create_task(
                send_any(state, state.unsent_messages.popleft())
            )
            return CheckResult.RESTART
        return CheckResult.CONTINUE

    if not state.send_task.done():
        return CheckResult.CONTINUE

    finished_task = state.send_task
    state.send_task = None
    finished_task.result()

    if state.unsent_messages:
        # save an event loop cycle
        state.send_task = asyncio.create_task(
            send_any(state, state.unsent_messages.popleft())
        )

    return CheckResult.RESTART



import asyncio
import base64
import hashlib
import io
import secrets
import tempfile
import time
from typing import TYPE_CHECKING, List, cast

import aiohttp
from lonelypsp.auth.config import AuthResult
from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.stateful.messages.configure import S2B_ConfigureParser
from lonelypsp.stateful.messages.confirm_configure import (
    B2S_ConfirmConfigure,
    serialize_b2s_confirm_configure,
)
from lonelypsp.stateful.parser_helpers import parse_s2b_message_prefix
from lonelypsp.util.bounded_deque import BoundedDeque
from lonelypsp.util.drainable_asyncio_queue import DrainableAsyncioQueue

from lonelypss.util.websocket_message import WSMessageBytes
from lonelypss.ws.handlers.protocol import StateHandler
from lonelypss.ws.simple_receiver import SimpleReceiver
from lonelypss.ws.state import (
    Compressor,
    CompressorPreparing,
    CompressorReady,
    CompressorState,
    CompressorTrainingDataCollector,
    CompressorTrainingInfoBeforeLowWatermark,
    CompressorTrainingInfoType,
    ConnectionConfiguration,
    State,
    StateClosing,
    StateOpen,
    StateType,
    StateWaitingConfigure,
)
from lonelypss.ws.util import make_websocket_read_task


async def _make_standard_compressor(state: StateWaitingConfigure) -> CompressorReady:
    return CompressorReady(
        type=CompressorState.READY,
        identifier=1,
        level=3,
        min_size=state.broadcaster_config.compression_trained_max_size,
        max_size=None,
        data=None,
        compressors=list(),
        decompressors=list(),
    )


async def _make_preset_compressor(
    state: StateWaitingConfigure, compressor_id: int
) -> CompressorReady:
    compressor_info = await state.broadcaster_config.get_compression_dictionary_by_id(
        compressor_id
    )
    if compressor_info is None:
        raise ValueError(f"Unknown compressor ID {compressor_id}")
    zdict, level = compressor_info

    return CompressorReady(
        type=CompressorState.READY,
        identifier=compressor_id,
        level=level,
        min_size=state.broadcaster_config.compression_min_size,
        max_size=None,
        data=zdict,
        compressors=list(),
        decompressors=list(),
    )


async def handle_waiting_configure(state: State) -> State:
    """Waits for a configure message to be sent over the websocket; if the subscriber
    sends anything else, moves to closing.
    """

    assert state.type == StateType.WAITING_CONFIGURE
    ws_message = await state.read_task
    if ws_message["type"] == "websocket.disconnect":
        return StateClosing(type=StateType.CLOSING, websocket=state.websocket)
    if ws_message["type"] != "websocket.receive":
        return StateClosing(type=StateType.CLOSING, websocket=state.websocket)
    if "bytes" not in ws_message:
        return StateClosing(type=StateType.CLOSING, websocket=state.websocket)

    raw_message = cast(WSMessageBytes, ws_message)
    raw_message_reader = io.BytesIO(raw_message["bytes"])
    prefix = parse_s2b_message_prefix(raw_message_reader)
    if prefix.type != SubscriberToBroadcasterStatefulMessageType.CONFIGURE:
        return StateClosing(type=StateType.CLOSING, websocket=state.websocket)

    message = S2B_ConfigureParser.parse(prefix.flags, prefix.type, raw_message_reader)
    auth_result = await state.broadcaster_config.is_stateful_configure_allowed(
        message=message, now=time.time()
    )
    if auth_result != AuthResult.OK:
        return StateClosing(type=StateType.CLOSING, websocket=state.websocket)

    receiver = SimpleReceiver()
    receiver_id = await state.internal_receiver.register_receiver(receiver)
    try:
        broadcaster_nonce = secrets.token_bytes(32)
        tracing = b""  # TODO: tracing
        authorization = (
            await state.broadcaster_config.authorize_stateful_confirm_configure(
                broadcaster_nonce=broadcaster_nonce, tracing=tracing, now=time.time()
            )
        )
        connection_nonce = hashlib.sha256(
            message.subscriber_nonce + broadcaster_nonce
        ).digest()

        compressors: List[Compressor] = []
        if state.broadcaster_config.compression_allowed and message.enable_zstd:
            compressors.append(
                CompressorPreparing(
                    type=CompressorState.PREPARING,
                    identifier=1,
                    task=asyncio.create_task(_make_standard_compressor(state)),
                )
            )

            if message.initial_dict > 1:
                compressors.append(
                    CompressorPreparing(
                        type=CompressorState.PREPARING,
                        identifier=message.initial_dict,
                        task=asyncio.create_task(
                            _make_preset_compressor(state, message.initial_dict)
                        ),
                    )
                )

        return StateOpen(
            type=StateType.OPEN,
            websocket=state.websocket,
            broadcaster_config=state.broadcaster_config,
            connection_config=ConnectionConfiguration(
                enable_zstd=message.enable_zstd,
                enable_training=message.enable_zstd and message.enable_training,
            ),
            nonce_b64=base64.b64encode(connection_nonce).decode("ascii"),
            internal_receiver=state.internal_receiver,
            my_receiver=receiver,
            my_receiver_id=receiver_id,
            client_session=aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(
                    total=state.broadcaster_config.outgoing_http_timeout_total,
                    connect=state.broadcaster_config.outgoing_http_timeout_connect,
                    sock_read=state.broadcaster_config.outgoing_http_timeout_sock_read,
                    sock_connect=state.broadcaster_config.outgoing_http_timeout_sock_connect,
                )
            ),
            compressors=compressors,
            compressor_training_info=(
                None
                if not state.broadcaster_config.allow_training
                or not message.enable_training
                else (
                    CompressorTrainingInfoBeforeLowWatermark(
                        type=CompressorTrainingInfoType.BEFORE_LOW_WATERMARK,
                        compressor_id=65536,
                        collector=CompressorTrainingDataCollector(
                            messages=0,
                            length=0,
                            tmpfile=tempfile.TemporaryFile(mode="w+b", buffering=-1),
                            pending=set(),
                        ),
                    )
                )
            ),
            broadcaster_counter=1,
            subscriber_counter=-1,
            read_task=make_websocket_read_task(state.websocket),
            notify_stream_state=None,
            send_task=asyncio.create_task(
                state.websocket.send_bytes(
                    serialize_b2s_confirm_configure(
                        B2S_ConfirmConfigure(
                            type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_CONFIGURE,
                            broadcaster_nonce=broadcaster_nonce,
                            authorization=authorization,
                            tracing=tracing,
                        ),
                        minimal_headers=state.broadcaster_config.websocket_minimal_headers,
                    )
                )
            ),
            process_task=None,
            unprocessed_messages=BoundedDeque(
                maxlen=state.broadcaster_config.websocket_max_unprocessed_receives
            ),
            unsent_messages=BoundedDeque(
                maxlen=state.broadcaster_config.websocket_max_pending_sends
            ),
            expecting_acks=DrainableAsyncioQueue(
                max_size=state.broadcaster_config.websocket_send_max_unacknowledged or 0
            ),
            backgrounded=set(),
        )
    except BaseException:
        await state.internal_receiver.unregister_receiver(receiver_id)
        raise


if TYPE_CHECKING:
    _: StateHandler = handle_waiting_configure



import asyncio
import io
import os
import time
from enum import Enum, auto
from tempfile import SpooledTemporaryFile
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Annotated,
    AsyncIterator,
    Literal,
    Optional,
    Type,
    Union,
    cast,
)

from fastapi import APIRouter, Header, Request
from fastapi.responses import Response
from lonelypsp.auth.config import AuthResult
from lonelypsp.compat import assert_never
from lonelypsp.stateless.constants import BroadcasterToSubscriberStatelessMessageType
from lonelypsp.stateless.make_strong_etag import (
    GlobAndRecovery,
    StrongEtag,
    TopicAndRecovery,
    create_strong_etag_generator,
)
from lonelypsp.sync_io import PreallocatedBytesIO

from lonelypss.config.set_subscriptions_info import SetSubscriptionsInfo
from lonelypss.middleware.config import get_config_from_request
from lonelypss.util.async_io import AsyncReadableBytesIO, async_read_exact
from lonelypss.util.request_body_io import AsyncIterableAIO
from lonelypss.util.sync_io import SyncIOBaseLikeIO, SyncReadableBytesIO, read_exact

router = APIRouter()


class _SubscriptionInfoFromStreamState(Enum):
    NOT_ENTERED = auto()
    TOPICS = auto()
    GLOBS_NOT_PEEKED = auto()
    GLOBS = auto()
    CLOSED = auto()


class _SubscriptionInfoFromStreamTopicsIter:
    def __init__(
        self,
        parent: "Union[_SubscriptionInfoFromSyncStream, _SubscriptionInfoFromAsyncStream]",
    ) -> None:
        self.parent = parent

    def __aiter__(self) -> "_SubscriptionInfoFromStreamTopicsIter":
        return self

    async def __anext__(self) -> TopicAndRecovery:
        assert self.parent.state == _SubscriptionInfoFromStreamState.TOPICS, "destroyed"
        return cast(TopicAndRecovery, await self.parent.__anext__())


class _SubscriptionInfoFromStreamGlobsIter:
    def __init__(
        self,
        parent: "Union[_SubscriptionInfoFromSyncStream, _SubscriptionInfoFromAsyncStream]",
    ) -> None:
        self.parent = parent

    def __aiter__(self) -> "_SubscriptionInfoFromStreamGlobsIter":
        return self

    async def __anext__(self) -> GlobAndRecovery:
        assert (
            self.parent.state == _SubscriptionInfoFromStreamState.GLOBS_NOT_PEEKED
            or self.parent.state == _SubscriptionInfoFromStreamState.GLOBS
        ), "destroyed"
        return cast(GlobAndRecovery, await self.parent.__anext__())


class _SubscriptionInfoFromSyncStream:
    def __init__(self, stream: SyncReadableBytesIO) -> None:
        self.stream = stream
        self.state: _SubscriptionInfoFromStreamState = (
            _SubscriptionInfoFromStreamState.NOT_ENTERED
        )
        self.remaining: int = 0

    def __enter__(self) -> "_SubscriptionInfoFromSyncStream":
        assert (
            self.state == _SubscriptionInfoFromStreamState.NOT_ENTERED
        ), "not reentrant"

        try:
            num_topics_bytes = read_exact(self.stream, 4)
            num_topics = int.from_bytes(num_topics_bytes, "big")
            self.remaining = num_topics
            self.state = _SubscriptionInfoFromStreamState.TOPICS
        except BaseException:
            self.state = _SubscriptionInfoFromStreamState.CLOSED
            raise

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        assert self.state != _SubscriptionInfoFromStreamState.NOT_ENTERED, "not entered"
        self.state = _SubscriptionInfoFromStreamState.CLOSED

    async def __anext__(self) -> Union[GlobAndRecovery, TopicAndRecovery]:
        if self.state == _SubscriptionInfoFromStreamState.TOPICS:
            if self.remaining <= 0:
                raise StopAsyncIteration

            try:
                topic_length_bytes = read_exact(self.stream, 2)
                topic_length = int.from_bytes(topic_length_bytes, "big")
                topic = read_exact(self.stream, topic_length)
                recovery_length_bytes = read_exact(self.stream, 2)
                recovery_length = int.from_bytes(recovery_length_bytes, "big")
                recovery_bytes = (
                    None
                    if recovery_length == 0
                    else read_exact(self.stream, recovery_length)
                )
                recovery_str = (
                    None if recovery_bytes is None else recovery_bytes.decode("utf-8")
                )
                self.remaining -= 1
            except BaseException:
                self.state = _SubscriptionInfoFromStreamState.CLOSED
                raise

            return TopicAndRecovery(topic, recovery_str)

        if self.state == _SubscriptionInfoFromStreamState.GLOBS_NOT_PEEKED:
            try:
                while self.remaining > 0:
                    topic_length_bytes = read_exact(self.stream, 2)
                    topic_length = int.from_bytes(topic_length_bytes, "big")
                    read_exact(self.stream, topic_length)
                    recovery_length_bytes = read_exact(self.stream, 2)
                    recovery_length = int.from_bytes(recovery_length_bytes, "big")
                    if recovery_length > 0:
                        read_exact(self.stream, recovery_length)
                    self.remaining -= 1
                    if self.remaining % 10 == 0:
                        await asyncio.sleep(0)

                num_globs_bytes = read_exact(self.stream, 4)
                num_globs = int.from_bytes(num_globs_bytes, "big")
                self.remaining = num_globs
                self.state = _SubscriptionInfoFromStreamState.GLOBS
            except BaseException:
                self.state = _SubscriptionInfoFromStreamState.CLOSED
                raise

        if self.state == _SubscriptionInfoFromStreamState.GLOBS:
            if self.remaining <= 0:
                raise StopAsyncIteration

            try:
                glob_length_bytes = read_exact(self.stream, 2)
                glob_length = int.from_bytes(glob_length_bytes, "big")
                glob_bytes = read_exact(self.stream, glob_length)
                glob = glob_bytes.decode("utf-8")
                recovery_length_bytes = read_exact(self.stream, 2)
                recovery_length = int.from_bytes(recovery_length_bytes, "big")
                recovery_bytes = (
                    None
                    if recovery_length == 0
                    else read_exact(self.stream, recovery_length)
                )
                recovery_str = (
                    None if recovery_bytes is None else recovery_bytes.decode("utf-8")
                )
                self.remaining -= 1
            except BaseException:
                self.state = _SubscriptionInfoFromStreamState.CLOSED
                raise

            return GlobAndRecovery(glob, recovery_str)

        raise AssertionError("not topics or globs")

    def __aiter__(self) -> "_SubscriptionInfoFromSyncStream":
        return self

    def topics(self) -> AsyncIterator[TopicAndRecovery]:
        assert self.state == _SubscriptionInfoFromStreamState.TOPICS, "not topics"
        return _SubscriptionInfoFromStreamTopicsIter(self)

    def globs(self) -> AsyncIterator[GlobAndRecovery]:
        if self.state == _SubscriptionInfoFromStreamState.TOPICS:
            self.state = _SubscriptionInfoFromStreamState.GLOBS_NOT_PEEKED

        if (
            self.state != _SubscriptionInfoFromStreamState.GLOBS_NOT_PEEKED
            and self.state != _SubscriptionInfoFromStreamState.GLOBS
        ):
            raise AssertionError("not globs")

        return _SubscriptionInfoFromStreamGlobsIter(self)


class _SubscriptionInfoFromAsyncStream:
    def __init__(self, stream: AsyncReadableBytesIO) -> None:
        self.stream = stream
        self.state: _SubscriptionInfoFromStreamState = (
            _SubscriptionInfoFromStreamState.NOT_ENTERED
        )
        self.remaining: int = 0

    async def __aenter__(self) -> "_SubscriptionInfoFromAsyncStream":
        assert (
            self.state == _SubscriptionInfoFromStreamState.NOT_ENTERED
        ), "not reentrant"

        try:
            num_topics_bytes = await async_read_exact(self.stream, 4)
            num_topics = int.from_bytes(num_topics_bytes, "big")
            self.remaining = num_topics
            self.state = _SubscriptionInfoFromStreamState.TOPICS
        except BaseException:
            self.state = _SubscriptionInfoFromStreamState.CLOSED
            raise

        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        assert self.state != _SubscriptionInfoFromStreamState.NOT_ENTERED, "not entered"
        self.state = _SubscriptionInfoFromStreamState.CLOSED

    async def __anext__(self) -> Union[GlobAndRecovery, TopicAndRecovery]:
        if self.state == _SubscriptionInfoFromStreamState.TOPICS:
            if self.remaining <= 0:
                raise StopAsyncIteration

            try:
                topic_length_bytes = await async_read_exact(self.stream, 2)
                topic_length = int.from_bytes(topic_length_bytes, "big")
                topic = await async_read_exact(self.stream, topic_length)
                recovery_length_bytes = await async_read_exact(self.stream, 2)
                recovery_length = int.from_bytes(recovery_length_bytes, "big")
                recovery_bytes = (
                    None
                    if recovery_length == 0
                    else await async_read_exact(self.stream, recovery_length)
                )
                recovery_str = (
                    None if recovery_bytes is None else recovery_bytes.decode("utf-8")
                )
                self.remaining -= 1
            except BaseException:
                self.state = _SubscriptionInfoFromStreamState.CLOSED
                raise

            return TopicAndRecovery(topic, recovery_str)

        if self.state == _SubscriptionInfoFromStreamState.GLOBS_NOT_PEEKED:
            try:
                while self.remaining > 0:
                    topic_length_bytes = await async_read_exact(self.stream, 2)
                    topic_length = int.from_bytes(topic_length_bytes, "big")
                    await async_read_exact(self.stream, topic_length)
                    recovery_length_bytes = await async_read_exact(self.stream, 2)
                    recovery_length = int.from_bytes(recovery_length_bytes, "big")
                    if recovery_length > 0:
                        await async_read_exact(self.stream, recovery_length)
                    self.remaining -= 1

                num_globs_bytes = await async_read_exact(self.stream, 4)
                num_globs = int.from_bytes(num_globs_bytes, "big")
                self.remaining = num_globs
                self.state = _SubscriptionInfoFromStreamState.GLOBS
            except BaseException:
                self.state = _SubscriptionInfoFromStreamState.CLOSED
                raise

        if self.state == _SubscriptionInfoFromStreamState.GLOBS:
            if self.remaining <= 0:
                raise StopAsyncIteration

            try:
                glob_length_bytes = await async_read_exact(self.stream, 2)
                glob_length = int.from_bytes(glob_length_bytes, "big")
                glob_bytes = await async_read_exact(self.stream, glob_length)
                glob = glob_bytes.decode("utf-8")
                recovery_length_bytes = await async_read_exact(self.stream, 2)
                recovery_length = int.from_bytes(recovery_length_bytes, "big")
                recovery_bytes = (
                    None
                    if recovery_length == 0
                    else await async_read_exact(self.stream, recovery_length)
                )
                recovery_str = (
                    None if recovery_bytes is None else recovery_bytes.decode("utf-8")
                )
                self.remaining -= 1
            except BaseException:
                self.state = _SubscriptionInfoFromStreamState.CLOSED
                raise

            return GlobAndRecovery(glob, recovery_str)

        raise AssertionError("not topics or globs")

    def __aiter__(self) -> "_SubscriptionInfoFromAsyncStream":
        return self

    def topics(self) -> AsyncIterator[TopicAndRecovery]:
        assert self.state == _SubscriptionInfoFromStreamState.TOPICS, "not topics"
        return _SubscriptionInfoFromStreamTopicsIter(self)

    def globs(self) -> AsyncIterator[GlobAndRecovery]:
        if self.state == _SubscriptionInfoFromStreamState.TOPICS:
            self.state = _SubscriptionInfoFromStreamState.GLOBS_NOT_PEEKED

        if (
            self.state != _SubscriptionInfoFromStreamState.GLOBS_NOT_PEEKED
            and self.state != _SubscriptionInfoFromStreamState.GLOBS
        ):
            raise AssertionError("not globs")

        return _SubscriptionInfoFromStreamGlobsIter(self)


class _CopyPassthroughStream:
    def __init__(self, source: AsyncReadableBytesIO, dest: SyncIOBaseLikeIO):
        self.source = source
        self.dest = dest

    async def read(self, n: int, /) -> bytes:
        data = await self.source.read(n)
        self.dest.write(data)
        return data


if TYPE_CHECKING:
    _a: Type[SetSubscriptionsInfo] = _SubscriptionInfoFromSyncStream
    _b: Type[SetSubscriptionsInfo] = _SubscriptionInfoFromAsyncStream
    _c: Type[AsyncIterator[Union[GlobAndRecovery, TopicAndRecovery]]] = (
        _SubscriptionInfoFromSyncStream
    )
    _d: Type[AsyncIterator[TopicAndRecovery]] = _SubscriptionInfoFromStreamTopicsIter
    _e: Type[AsyncIterator[GlobAndRecovery]] = _SubscriptionInfoFromStreamGlobsIter
    _f: Type[AsyncReadableBytesIO] = _CopyPassthroughStream


@router.post(
    "/v1/set_subscriptions",
    status_code=200,
    responses={
        401: {"description": "Authorization header is required but not provided"},
        403: {"description": "Authorization header is provided but invalid"},
        500: {"description": "Unexpected error occurred"},
        503: {"description": "Service is unavailable, try again soon"},
    },
)
async def check_subscriptions(
    request: Request, authorization: Annotated[Optional[str], Header()] = None
) -> Response:
    """Sets the subscriptions associated with the URL to a specific value. This
    is an idempotent variant of the subscribe/unsubscribe endpoints. Most of the
    time, it is easier, more efficient, and more resilient to use this endpoint
    instead of mutating the subscriptions in-place. The exceptions are:

    - in a websocket, since each websocket is self-contained, the subscribe/unsubscribe
      options are sufficient

    - if there are actually a large number of partial changes (especially with a large
      number of subscriptions) it may be more efficient to use the subscribe/unsubscribe
      endpoints

    - if there are an extremely large number of subscriptions, it may be preferable
      to use this to clear out the subscriptions, then add them via separate
      subscribe calls to allow for resumable processing

    NOTE: this is not atomic but there are some guarrantees; see the lonelypsp
    documentation for stateless `SET_SUBSCRIPTION`

    ### request body
    - 2 bytes (T): length of tracing data, big-endian, unsigned
    - T bytes: the tracing data
    - 2 bytes (N): length of the subscriber url to set, big-endian, unsigned
    - N bytes: the url to set, utf-8 encoded
    - 1 byte (reserved for etag format): 0
    - 64 bytes: the strong etag, will be rechecked
    - 4 bytes (E): the number of exact topics to set, big-endian, unsigned
    - REPEAT E TIMES: (in ascending lexicographic order)
      - 2 bytes (L): length of the topic, big-endian, unsigned
      - L bytes: the topic
    - 4 bytes (G): the number of glob patterns to set, big-endian, unsigned
    - REPEAT G TIMES: (in ascending lexicographic order)
      - 2 bytes (L): length of the glob pattern, big-endian, unsigned
      - L bytes: the glob pattern, utf-8 encoded

    ### response body
    - 2 bytes (type): int(RESPONSE_GENERIC), big endian, unsigned
    - 2 bytes (A): big-endian, unsigned, the length of the authorization
    - A bytes: the authorization
    - 2 bytes (T): big-endian, unsigned, the length of tracing data
    - T bytes: the tracing data
    """
    config = get_config_from_request(request)

    try:
        with SpooledTemporaryFile(
            max_size=config.message_body_spool_size, mode="w+b"
        ) as subs_info_buffer:
            stream = request.stream()

            try:
                body = AsyncIterableAIO(stream.__aiter__())

                tracing_length_bytes = await async_read_exact(body, 2)
                tracing_length = int.from_bytes(tracing_length_bytes, "big")
                tracing = await async_read_exact(body, tracing_length)

                url_length_bytes = await async_read_exact(body, 2)
                url_length = int.from_bytes(url_length_bytes, "big")
                url_bytes = await async_read_exact(body, url_length)
                url = url_bytes.decode("utf-8")

                if (await async_read_exact(body, 1))[0] != 0:
                    return Response(status_code=400)
                etag_format: Literal[0] = 0

                etag = await async_read_exact(body, 64)

                auth_at = time.time()

                # note how we allow for checking some or all of the authorization
                # header before reading an arbitrary amount of data from the body
                # (at most we've read 2 + 64kibibytes + 1 + 64 bytes so far)
                cp_stream = _CopyPassthroughStream(body, subs_info_buffer)
                async with _SubscriptionInfoFromAsyncStream(
                    cp_stream
                ) as subscription_info:
                    auth_result = await config.is_set_subscriptions_allowed(
                        tracing=tracing,
                        url=url,
                        strong_etag=StrongEtag(format=0, etag=etag),
                        subscriptions=subscription_info,
                        now=auth_at,
                        authorization=authorization,
                    )

                if auth_result == AuthResult.UNAUTHORIZED:
                    return Response(status_code=401)
                elif auth_result == AuthResult.FORBIDDEN:
                    return Response(status_code=403)
                elif auth_result == AuthResult.UNAVAILABLE:
                    return Response(status_code=503)
                elif auth_result != AuthResult.OK:
                    assert_never(auth_result)

                while True:
                    chunk = await cp_stream.read(io.DEFAULT_BUFFER_SIZE)
                    if not chunk:
                        break
            finally:
                await stream.aclose()

            # check etag was actually correct
            subs_info_buffer.seek(0, os.SEEK_SET)
            topics_gen = create_strong_etag_generator(url, recheck_sort=True)
            with _SubscriptionInfoFromSyncStream(subs_info_buffer) as subscription_info:
                async for topic in subscription_info.topics():
                    topics_gen.add_topic(topic)

                glob_gen = topics_gen.finish_topics()
                del topics_gen
                async for glob in subscription_info.globs():
                    glob_gen.add_glob(glob)

                real_etag = glob_gen.finish()

            if real_etag.format != etag_format or real_etag.etag != etag:
                return Response(status_code=400)

            # actually set the subscriptions
            subs_info_buffer.seek(0, os.SEEK_SET)
            with _SubscriptionInfoFromSyncStream(subs_info_buffer) as subscription_info:
                set_result = await config.set_subscriptions(
                    url=url, strong_etag=real_etag, subscriptions=subscription_info
                )

            if set_result == "unavailable":
                return Response(status_code=503)
            if set_result != "success":
                assert_never(set_result)

            resp_tracing = b""  # TODO: tracing
            resp_authorization = await config.authorize_set_subscriptions_response(
                tracing=resp_tracing, strong_etag=real_etag, now=time.time()
            )
            resp_authorization_bytes = (
                b""
                if resp_authorization is None
                else resp_authorization.encode("utf-8")
            )
            resp_body = PreallocatedBytesIO(
                2 + 2 + len(resp_authorization_bytes) + 2 + len(resp_tracing)
            )
            resp_body.write(
                int(
                    BroadcasterToSubscriberStatelessMessageType.RESPONSE_GENERIC
                ).to_bytes(2, "big")
            )
            resp_body.write(len(resp_authorization_bytes).to_bytes(2, "big"))
            resp_body.write(resp_authorization_bytes)
            resp_body.write(len(resp_tracing).to_bytes(2, "big"))
            resp_body.write(resp_tracing)
            return Response(
                status_code=200,
                headers={"Content-Type": "application/octet-stream"},
                content=memoryview(resp_body.buffer),
            )

    except ValueError:
        return Response(status_code=400)
